# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""

import binascii
import pprint
import re
from abc import ABC
from collections import deque
from typing import (
    Any,
    Deque,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from aea.protocols.base import Message
from web3.types import Nonce, TxData, Wei

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviour_utils import RPCResponseStatus
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.common import (
    RandomnessBehaviour,
    SelectKeeperBehaviour,
)
from packages.valory.skills.abstract_round_abci.utils import VerifyDrand
from packages.valory.skills.transaction_settlement_abci.models import TransactionParams
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    skill_input_hex_to_payload,
    tx_hist_payload_to_hex,
)
from packages.valory.skills.transaction_settlement_abci.payloads import (
    CheckTransactionHistoryPayload,
    FinalizationTxPayload,
    RandomnessPayload,
    ResetPayload,
    SelectKeeperPayload,
    SignaturePayload,
    SynchronizeLateMessagesPayload,
    ValidatePayload,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    CheckLateTxHashesRound,
    CheckTransactionHistoryRound,
    CollectSignatureRound,
    FinalizationRound,
    RandomnessTransactionSubmissionRound,
    ResetRound,
    SelectKeeperTransactionSubmissionARound,
    SelectKeeperTransactionSubmissionBAfterTimeoutRound,
    SelectKeeperTransactionSubmissionBRound,
    SynchronizeLateMessagesRound,
    SynchronizedData,
    TransactionSubmissionAbciApp,
    ValidateTransactionRound,
)


TxDataType = Dict[str, Union[VerificationStatus, Deque[str], int, Set[str], str]]


drand_check = VerifyDrand()

REVERT_CODE_RE = r"\s(GS\d{3})[^\d]"

# This mapping was copied from:
# https://github.com/safe-global/safe-contracts/blob/ce5cbd256bf7a8a34538c7e5f1f2366a9d685f34/docs/error_codes.md
REVERT_CODES_TO_REASONS: Dict[str, str] = {
    "GS000": "Could not finish initialization",
    "GS001": "Threshold needs to be defined",
    "GS010": "Not enough gas to execute Safe transaction",
    "GS011": "Could not pay gas costs with ether",
    "GS012": "Could not pay gas costs with token",
    "GS013": "Safe transaction failed when gasPrice and safeTxGas were 0",
    "GS020": "Signatures data too short",
    "GS021": "Invalid contract signature location: inside static part",
    "GS022": "Invalid contract signature location: length not present",
    "GS023": "Invalid contract signature location: data not complete",
    "GS024": "Invalid contract signature provided",
    "GS025": "Hash has not been approved",
    "GS026": "Invalid owner provided",
    "GS030": "Only owners can approve a hash",
    "GS031": "Method can only be called from this contract",
    "GS100": "Modules have already been initialized",
    "GS101": "Invalid module address provided",
    "GS102": "Module has already been added",
    "GS103": "Invalid prevModule, module pair provided",
    "GS104": "Method can only be called from an enabled module",
    "GS200": "Owners have already been setup",
    "GS201": "Threshold cannot exceed owner count",
    "GS202": "Threshold needs to be greater than 0",
    "GS203": "Invalid owner address provided",
    "GS204": "Address is already an owner",
    "GS205": "Invalid prevOwner, owner pair provided",
    "GS300": "Guard does not implement IERC165",
}


class TransactionSettlementBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the common apps' skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> TransactionParams:
        """Return the params."""
        return cast(TransactionParams, super().params)

    @staticmethod
    def serialized_keepers(keepers: Deque[str], keeper_retries: int) -> str:
        """Get the keepers serialized."""
        if len(keepers) == 0:
            return ""
        keepers_ = "".join(keepers)
        keeper_retries_ = keeper_retries.to_bytes(32, "big").hex()
        concatenated = keeper_retries_ + keepers_

        return concatenated

    def get_gas_price_params(self, tx_body: dict) -> List[str]:
        """Guess the gas strategy from the transaction params"""
        strategy_to_params: Dict[str, List[str]] = {
            "eip": ["maxPriorityFeePerGas", "maxFeePerGas"],
            "gas_station": ["gasPrice"],
        }

        for strategy, params in strategy_to_params.items():
            if all(param in tx_body for param in params):
                self.context.logger.info(f"Detected gas strategy: {strategy}")
                return params

        return []

    def _get_tx_data(
        self, message: ContractApiMessage, use_flashbots: bool
    ) -> Generator[None, None, TxDataType]:
        """Get the transaction data from a `ContractApiMessage`."""
        tx_data: TxDataType = {
            "status": VerificationStatus.PENDING,
            "keepers": self.synchronized_data.keepers,
            "keeper_retries": self.synchronized_data.keeper_retries,
            "blacklisted_keepers": self.synchronized_data.blacklisted_keepers,
            "tx_digest": "",
        }

        # Check for errors in the transaction preparation
        if (
            message.performative == ContractApiMessage.Performative.ERROR
            and message.message is not None
        ):
            if self._safe_nonce_reused(message.message):
                tx_data["status"] = VerificationStatus.VERIFIED
            else:
                tx_data["status"] = VerificationStatus.ERROR
            self.context.logger.warning(self._parse_revert_reason(message))
            return tx_data

        # Check that we have a RAW_TRANSACTION response
        if message.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
            self.context.logger.warning(
                f"get_raw_safe_transaction unsuccessful! Received: {message}"
            )
            return tx_data

        # Send transaction
        tx_digest, rpc_status = yield from self.send_raw_transaction(
            message.raw_transaction, use_flashbots
        )

        # Handle transaction results
        if rpc_status == RPCResponseStatus.ALREADY_KNOWN:
            self.context.logger.warning(
                "send_raw_transaction unsuccessful! Transaction is already in the mempool! Will attempt to verify it."
            )

        if rpc_status == RPCResponseStatus.INCORRECT_NONCE:
            tx_data["status"] = VerificationStatus.ERROR
            self.context.logger.warning(
                "send_raw_transaction unsuccessful! Incorrect nonce."
            )

        if rpc_status == RPCResponseStatus.INSUFFICIENT_FUNDS:
            # blacklist self.
            tx_data["status"] = VerificationStatus.INSUFFICIENT_FUNDS
            blacklisted = cast(Deque[str], tx_data["keepers"]).popleft()
            tx_data["keeper_retries"] = 1
            cast(Set[str], tx_data["blacklisted_keepers"]).add(blacklisted)
            self.context.logger.warning(
                "send_raw_transaction unsuccessful! Insufficient funds."
            )

        if rpc_status not in {
            RPCResponseStatus.SUCCESS,
            RPCResponseStatus.ALREADY_KNOWN,
        }:
            self.context.logger.warning(
                f"send_raw_transaction unsuccessful! Received: {rpc_status}"
            )
            return tx_data

        tx_data["tx_digest"] = cast(str, tx_digest)

        nonce = Nonce(int(cast(str, message.raw_transaction.body["nonce"])))
        fallback_gas = message.raw_transaction.body["gas"]

        # Get the gas params
        gas_price_params = self.get_gas_price_params(message.raw_transaction.body)

        gas_price = {
            gas_price_param: Wei(
                int(
                    cast(
                        str,
                        message.raw_transaction.body[gas_price_param],
                    )
                )
            )
            for gas_price_param in gas_price_params
        }

        # Set hash, nonce and tip.
        self.params.mutable_params.tx_hash = cast(str, tx_data["tx_digest"])
        if nonce == self.params.mutable_params.nonce:
            self.context.logger.info(
                "Attempting to replace transaction "
                f"with old gas price parameters {self.params.mutable_params.gas_price}, using new gas price parameters {gas_price}"
            )
        else:
            self.context.logger.info(
                f"Sent transaction for mining with gas parameters {gas_price}"
            )
            self.params.mutable_params.nonce = nonce
        self.params.mutable_params.gas_price = gas_price
        self.params.mutable_params.fallback_gas = fallback_gas

        return tx_data

    def _verify_tx(self, tx_hash: str) -> Generator[None, None, ContractApiMessage]:
        """Verify a transaction."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="verify_tx",
            tx_hash=tx_hash,
            owners=tuple(self.synchronized_data.participants),
            to_address=tx_params["to_address"],
            value=tx_params["ether_value"],
            data=tx_params["data"],
            safe_tx_gas=tx_params["safe_tx_gas"],
            signatures_by_owner={
                key: payload.signature
                for key, payload in self.synchronized_data.participant_to_signature.items()
            },
            operation=tx_params["operation"],
        )

        return contract_api_msg

    @staticmethod
    def _safe_nonce_reused(revert_reason: str) -> bool:
        """Check for GS026."""
        return "GS026" in revert_reason

    @staticmethod
    def _parse_revert_reason(message: ContractApiMessage) -> str:
        """Parse a revert reason and log a relevant message."""
        default_message = f"get_raw_safe_transaction unsuccessful! Received: {message}"

        revert_reason = message.message
        if not revert_reason:
            return default_message

        revert_match = re.findall(REVERT_CODE_RE, revert_reason)
        if revert_match is None or len(revert_match) != 1:
            return default_message

        revert_code = revert_match.pop()
        revert_explanation = REVERT_CODES_TO_REASONS.get(revert_code, None)
        if revert_explanation is None:
            return default_message

        return f"Received a {revert_code} revert error: {revert_explanation}."


class RandomnessTransactionSubmissionBehaviour(RandomnessBehaviour):
    """Retrieve randomness."""

    matching_round = RandomnessTransactionSubmissionRound
    payload_class = RandomnessPayload


class SelectKeeperTransactionSubmissionBehaviourA(  # pylint: disable=too-many-ancestors
    SelectKeeperBehaviour, TransactionSettlementBaseBehaviour
):
    """Select the keeper agent."""

    matching_round = SelectKeeperTransactionSubmissionARound
    payload_class = SelectKeeperPayload

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = deque((self._select_keeper(),))
            payload = self.payload_class(
                self.context.agent_address, self.serialized_keepers(keepers, 1)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class SelectKeeperTransactionSubmissionBehaviourB(  # pylint: disable=too-many-ancestors
    SelectKeeperTransactionSubmissionBehaviourA
):
    """Select the keeper b agent."""

    matching_round = SelectKeeperTransactionSubmissionBRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
            - If we have not selected enough keepers for the period,
                select a keeper randomly and add it to the keepers' queue, with top priority.
            - Otherwise, cycle through the keepers' subset, using the following logic:
                A `PENDING` verification status means that we have not received any errors,
                therefore, all we know is that the tx has not been mined yet due to low pricing.
                Consequently, we are going to retry with the same keeper in order to replace the transaction.
                However, if we receive a status other than `PENDING`, we need to cycle through the keepers' subset.
                Moreover, if the current keeper has reached the allowed number of retries, then we cycle anyway.
            - Send the transaction with the keepers and wait for it to be mined.
            - Wait until ABCI application transitions to the next round.
            - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = self.synchronized_data.keepers
            keeper_retries = 1

            if (
                self.synchronized_data.keepers_threshold_exceeded
            ):  # TODO: I think this should be second prio
                keepers.rotate(-1)
                self.context.logger.info(f"Rotated keepers to: {keepers}.")
            elif (
                self.synchronized_data.keeper_retries
                != self.params.keeper_allowed_retries
                and self.synchronized_data.final_verification_status
                == VerificationStatus.PENDING
            ):
                keeper_retries += self.synchronized_data.keeper_retries
                self.context.logger.info(
                    f"Kept keepers and incremented retries: {keepers}."
                )
            else:
                keepers.appendleft(self._select_keeper())

            payload = self.payload_class(
                self.context.agent_address,
                self.serialized_keepers(keepers, keeper_retries),
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class SelectKeeperTransactionSubmissionBehaviourBAfterTimeout(  # pylint: disable=too-many-ancestors
    SelectKeeperTransactionSubmissionBehaviourB
):
    """Select the keeper b agent after a timeout."""

    matching_round = SelectKeeperTransactionSubmissionBAfterTimeoutRound


class ValidateTransactionBehaviour(TransactionSettlementBaseBehaviour):
    """Validate a transaction."""

    matching_round = ValidateTransactionRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Validate that the transaction hash provided by the keeper points to a
          valid transaction.
        - Send the transaction with the validation result and wait for it to be
          mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            is_correct = yield from self.has_transaction_been_sent()
            if is_correct:
                self.context.logger.info(
                    f"Finalized with transaction hash: {self.synchronized_data.to_be_validated_tx_hash}"
                )
            payload = ValidatePayload(self.context.agent_address, is_correct)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def has_transaction_been_sent(self) -> Generator[None, None, Optional[bool]]:
        """Transaction verification."""

        to_be_validated_tx_hash = self.synchronized_data.to_be_validated_tx_hash

        response = yield from self.get_transaction_receipt(
            to_be_validated_tx_hash,
            self.params.retry_timeout,
            self.params.retry_attempts,
        )
        if response is None:  # pragma: nocover
            self.context.logger.error(
                f"tx {to_be_validated_tx_hash} receipt check timed out!"
            )
            return None

        contract_api_msg = yield from self._verify_tx(to_be_validated_tx_hash)
        if (
            contract_api_msg.performative != ContractApiMessage.Performative.STATE
        ):  # pragma: nocover
            self.context.logger.error(
                f"verify_tx unsuccessful! Received: {contract_api_msg}"
            )
            return False

        verified = cast(bool, contract_api_msg.state.body["verified"])
        verified_log = (
            f"Verified result: {verified}"
            if verified
            else f"Verified result: {verified}, all: {contract_api_msg.state.body}"
        )
        self.context.logger.info(verified_log)
        return verified


class CheckTransactionHistoryBehaviour(TransactionSettlementBaseBehaviour):
    """Check the transaction history."""

    matching_round = CheckTransactionHistoryRound
    check_expected_to_be_verified = "The next tx check"

    @property
    def history(self) -> List[str]:
        """Get the history of hashes."""
        return self.synchronized_data.tx_hashes_history

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            verification_status, tx_hash = yield from self._check_tx_history()
            if verification_status == VerificationStatus.VERIFIED:
                msg = f"A previous transaction {tx_hash} has already been verified "
                msg += (
                    f"for {self.synchronized_data.to_be_validated_tx_hash}."
                    if self.synchronized_data.tx_hashes_history
                    else "and was synced after the finalization round timed out."
                )
                self.context.logger.info(msg)
            elif verification_status == VerificationStatus.NOT_VERIFIED:
                self.context.logger.info(
                    f"No previous transaction has been verified for "
                    f"{self.synchronized_data.to_be_validated_tx_hash}."
                )

            verified_res = tx_hist_payload_to_hex(verification_status, tx_hash)
            payload = CheckTransactionHistoryPayload(
                self.context.agent_address, verified_res
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def _check_tx_history(
        self,
    ) -> Generator[None, None, Tuple[VerificationStatus, Optional[str]]]:
        """Check the transaction history."""
        if not self.history:
            self.context.logger.error(
                "An unexpected error occurred! The synchronized data do not contain any transaction hashes, "
                f"but entered the `{self.behaviour_id}` behaviour."
            )
            return VerificationStatus.ERROR, None

        self.context.logger.info(
            f"Starting check for the transaction history: {self.history}."
        )
        for tx_hash in self.history[::-1]:
            self.context.logger.info(f"Checking hash {tx_hash}...")
            contract_api_msg = yield from self._verify_tx(tx_hash)

            if (
                contract_api_msg.performative != ContractApiMessage.Performative.STATE
            ):  # pragma: nocover
                self.context.logger.error(
                    f"verify_tx unsuccessful for {tx_hash}! Received: {contract_api_msg}"
                )
                return VerificationStatus.ERROR, tx_hash

            verified = cast(bool, contract_api_msg.state.body["verified"])
            verified_log = f"Verified result for {tx_hash}: {verified}"

            if verified:
                self.context.logger.info(verified_log)
                return VerificationStatus.VERIFIED, tx_hash

            self.context.logger.info(
                verified_log + f", all: {contract_api_msg.state.body}"
            )

            status = cast(int, contract_api_msg.state.body["status"])
            if status == -1:
                self.context.logger.info(f"Tx hash {tx_hash} has no receipt!")
                # this loop might take a long time
                # we do not want to starve the rest of the behaviour
                # we yield which freezes this loop here until the
                # AbstractRoundBehaviour it belongs to, sends a tick to it
                yield
                continue

            tx_data = cast(TxData, contract_api_msg.state.body["transaction"])
            revert_reason = yield from self._get_revert_reason(tx_data)

            if revert_reason is not None:
                if self._safe_nonce_reused(revert_reason):
                    self.context.logger.info(
                        f"The safe's nonce has been reused for {tx_hash}. "
                        f"{self.check_expected_to_be_verified} is expected to be verified!"
                    )
                    # this loop might take a long time
                    # we do not want to starve the rest of the behaviour
                    # we yield which freezes this loop here until the
                    # AbstractRoundBehaviour it belongs to, sends a tick to it
                    yield
                    continue

                self.context.logger.warning(
                    f"Payload is invalid for {tx_hash}! Cannot continue. Received: {revert_reason}"
                )

            return VerificationStatus.INVALID_PAYLOAD, tx_hash

        return VerificationStatus.NOT_VERIFIED, None

    def _get_revert_reason(self, tx: TxData) -> Generator[None, None, Optional[str]]:
        """Get the revert reason of the given transaction."""
        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="revert_reason",
            tx=tx,
        )

        if (
            contract_api_msg.performative != ContractApiMessage.Performative.STATE
        ):  # pragma: nocover
            self.context.logger.error(
                f"An unexpected error occurred while checking {tx['hash'].hex()}: {contract_api_msg}"
            )
            return None

        return cast(str, contract_api_msg.state.body["revert_reason"])


class CheckLateTxHashesBehaviour(  # pylint: disable=too-many-ancestors
    CheckTransactionHistoryBehaviour
):
    """Check the late-arriving transaction hashes."""

    matching_round = CheckLateTxHashesRound
    check_expected_to_be_verified = "One of the next tx checks"

    @property
    def history(self) -> List[str]:
        """Get the history of hashes."""
        return [
            hash_
            for hashes in self.synchronized_data.late_arriving_tx_hashes.values()
            for hash_ in hashes
        ]


class SynchronizeLateMessagesBehaviour(TransactionSettlementBaseBehaviour):
    """Synchronize late-arriving messages behaviour."""

    matching_round = SynchronizeLateMessagesRound

    def __init__(self, **kwargs: Any):
        """Initialize a `SynchronizeLateMessagesBehaviour`"""
        super().__init__(**kwargs)
        # if we timed out during finalization, but we managed to receive a tx hash,
        # then we sync it here by initializing the `_tx_hashes` with the unsynced hash.
        self._tx_hashes: str = self.params.mutable_params.tx_hash
        self._messages_iterator: Iterator[ContractApiMessage] = iter(
            self.params.mutable_params.late_messages
        )
        self.use_flashbots = False

    def setup(self) -> None:
        """Setup the `SynchronizeLateMessagesBehaviour`."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )
        self.use_flashbots = tx_params["use_flashbots"]

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            current_message = next(self._messages_iterator, None)
            if current_message is not None:
                tx_data = yield from self._get_tx_data(
                    current_message, self.use_flashbots
                )
                self.context.logger.info(
                    f"Found a late arriving message {current_message}. Result data: {tx_data}"
                )
                # here, we concatenate the tx_hashes of all the late-arriving messages. Later, we will parse them.
                self._tx_hashes += cast(str, tx_data["tx_digest"])
                return

            payload = SynchronizeLateMessagesPayload(
                self.context.agent_address, self._tx_hashes
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            # reset the local parameters if we were able to send them.
            self.params.mutable_params.tx_hash = ""
            self.params.mutable_params.late_messages = []
            yield from self.wait_until_round_end()

        self.set_done()


class SignatureBehaviour(TransactionSettlementBaseBehaviour):
    """Signature behaviour."""

    matching_round = CollectSignatureRound

    def async_act(self) -> Generator:
        """
        Do the action.

        Steps:
        - Request the signature of the transaction hash.
        - Send the signature as a transaction and wait for it to be mined.
        - Wait until ABCI application transitions to the next round.
        - Go to the next behaviour (set done event).
        """

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            self.context.logger.info(
                f"Agreement reached on tx data: {self.synchronized_data.most_voted_tx_hash}"
            )
            signature_hex = yield from self._get_safe_tx_signature()
            payload = SignaturePayload(self.context.agent_address, signature_hex)

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()

    def _get_safe_tx_signature(self) -> Generator[None, None, str]:
        """Get signature of safe transaction hash."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )
        # is_deprecated_mode=True because we want to call Account.signHash,
        # which is the same used by gnosis-py
        safe_tx_hash_bytes = binascii.unhexlify(tx_params["safe_tx_hash"])
        signature_hex = yield from self.get_signature(
            safe_tx_hash_bytes, is_deprecated_mode=True
        )
        # remove the leading '0x'
        signature_hex = signature_hex[2:]
        self.context.logger.info(f"Signature: {signature_hex}")
        return signature_hex


class FinalizeBehaviour(TransactionSettlementBaseBehaviour):
    """Finalize behaviour."""

    matching_round = FinalizationRound

    def _i_am_not_sending(self) -> bool:
        """Indicates if the current agent is the sender or not."""
        return (
            self.context.agent_address
            != self.synchronized_data.most_voted_keeper_address
        )

    def async_act(self) -> Generator[None, None, None]:
        """
        Do the action.

        Steps:
        - If the agent is the keeper, then prepare the transaction and send it.
        - Otherwise, wait until the next round.
        - If a timeout is hit, set exit A event, otherwise set done event.
        """
        if self._i_am_not_sending():
            yield from self._not_sender_act()
        else:
            yield from self._sender_act()

    def _not_sender_act(self) -> Generator:
        """Do the non-sender action."""
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            self.context.logger.info(
                f"Waiting for the keeper to do its keeping: {self.synchronized_data.most_voted_keeper_address}"
            )
            yield from self.wait_until_round_end()
        self.set_done()

    def _sender_act(self) -> Generator[None, None, None]:
        """Do the sender action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            self.context.logger.info(
                "I am the designated sender, attempting to send the safe transaction..."
            )
            tx_data = yield from self._send_safe_transaction()
            if (
                tx_data["tx_digest"] == ""
                and tx_data["status"] == VerificationStatus.PENDING
            ) or tx_data["status"] == VerificationStatus.ERROR:
                self.context.logger.error(
                    "Did not succeed with finalising the transaction!"
                )
            elif tx_data["status"] == VerificationStatus.VERIFIED:
                self.context.logger.error(
                    "Trying to finalize a transaction which has been verified already!"
                )
            else:  # pragma: no cover
                self.context.logger.info(
                    f"Finalization tx digest: {cast(str, tx_data['tx_digest'])}"
                )
                self.context.logger.debug(
                    f"Signatures: {pprint.pformat(self.synchronized_data.participant_to_signature)}"
                )

            tx_hashes_history = self.synchronized_data.tx_hashes_history

            if tx_data["tx_digest"] != "":
                tx_hashes_history.append(cast(str, tx_data["tx_digest"]))

            tx_data_serialized = {
                "status_value": cast(VerificationStatus, tx_data["status"]).value,
                "serialized_keepers": self.serialized_keepers(
                    cast(Deque[str], tx_data["keepers"]),
                    cast(int, tx_data["keeper_retries"]),
                ),
                "blacklisted_keepers": "".join(
                    cast(Set[str], tx_data["blacklisted_keepers"])
                ),
                "tx_hashes_history": "".join(tx_hashes_history),
                "received_hash": bool(tx_data["tx_digest"]),
            }

            payload = FinalizationTxPayload(
                self.context.agent_address,
                cast(Dict[str, Union[str, int, bool]], tx_data_serialized),
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            # reset the local tx hash parameter if we were able to send it
            self.params.mutable_params.tx_hash = ""
            yield from self.wait_until_round_end()

        self.set_done()

    def _send_safe_transaction(
        self,
    ) -> Generator[None, None, TxDataType]:
        """Send a Safe transaction using the participants' signatures."""
        tx_params = skill_input_hex_to_payload(
            self.synchronized_data.most_voted_tx_hash
        )

        contract_api_msg = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(GnosisSafeContract.contract_id),
            contract_callable="get_raw_safe_transaction",
            sender_address=self.context.agent_address,
            owners=tuple(self.synchronized_data.participants),
            to_address=tx_params["to_address"],
            value=tx_params["ether_value"],
            data=tx_params["data"],
            safe_tx_gas=tx_params["safe_tx_gas"],
            signatures_by_owner={
                key: payload.signature
                for key, payload in self.synchronized_data.participant_to_signature.items()
            },
            nonce=self.params.mutable_params.nonce,
            old_price=self.params.mutable_params.gas_price,
            operation=tx_params["operation"],
            fallback_gas=self.params.mutable_params.fallback_gas,
            gas_price=self.params.gas_params.gas_price,
            max_fee_per_gas=self.params.gas_params.max_fee_per_gas,
            max_priority_fee_per_gas=self.params.gas_params.max_priority_fee_per_gas,
        )

        tx_data = yield from self._get_tx_data(
            contract_api_msg, tx_params["use_flashbots"]
        )
        return tx_data

    def handle_late_messages(self, behaviour_id: str, message: Message) -> None:
        """Store a potentially late-arriving message locally.

        :param behaviour_id: the id of the behaviour in which the message belongs to.
        :param message: the late arriving message to handle.
        """
        if (
            isinstance(message, ContractApiMessage)
            and behaviour_id == self.behaviour_id
        ):
            self.context.logger.info(f"Late message arrived: {message}")
            self.params.mutable_params.late_messages.append(message)
        else:
            super().handle_late_messages(behaviour_id, message)


class ResetBehaviour(TransactionSettlementBaseBehaviour):
    """Reset behaviour."""

    matching_round = ResetRound

    def async_act(self) -> Generator:
        """Do the action."""
        self.context.logger.info(
            f"Period {self.synchronized_data.period_count} was not finished. Resetting!"
        )
        payload = ResetPayload(
            self.context.agent_address, self.synchronized_data.period_count
        )
        yield from self.send_a2a_transaction(payload)
        yield from self.wait_until_round_end()
        self.set_done()


class TransactionSettlementRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the basic transaction settlement."""

    initial_behaviour_cls = RandomnessTransactionSubmissionBehaviour
    abci_app_cls = TransactionSubmissionAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        RandomnessTransactionSubmissionBehaviour,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourA,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourB,  # type: ignore
        SelectKeeperTransactionSubmissionBehaviourBAfterTimeout,  # type: ignore
        ValidateTransactionBehaviour,  # type: ignore
        CheckTransactionHistoryBehaviour,  # type: ignore
        SignatureBehaviour,  # type: ignore
        FinalizeBehaviour,  # type: ignore
        SynchronizeLateMessagesBehaviour,  # type: ignore
        CheckLateTxHashesBehaviour,  # type: ignore
        ResetBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the base round classes."""

# pylint: skip-file

import re
from enum import Enum
from typing import FrozenSet, List, Optional, Tuple, Union, cast

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    BaseSynchronizedData,
    BaseTxPayload,
    TransactionNotValidError,
)
from packages.valory.skills.abstract_round_abci.test_tools.rounds import (
    BaseOnlyKeeperSendsRoundTest,
    DummyCollectDifferentUntilAllRound,
    DummyCollectDifferentUntilThresholdRound,
    DummyCollectNonEmptyUntilThresholdRound,
    DummyCollectSameUntilAllRound,
    DummyCollectSameUntilThresholdRound,
    DummyCollectionRound,
    DummyEvent,
    DummyOnlyKeeperSendsRound,
    DummyTxPayload,
    DummyVotingRound,
    MAX_PARTICIPANTS,
    _BaseRoundTestClass,
    get_dummy_tx_payloads,
)


class TestCollectionRound(_BaseRoundTestClass):
    """Test class for CollectionRound."""

    def setup(
        self,
    ) -> None:
        """Setup test."""
        super().setup()

        self.test_round = DummyCollectionRound(synchronized_data=self.synchronized_data)

    def test_serialized_collection(self) -> None:
        """Test `serialized_collection` property."""
        assert self.test_round.serialized_collection == {}

        for payload in self.tx_payloads:
            self.test_round.process_payload(payload)

        mcs_key = "packages.valory.skills.abstract_round_abci.test_tools.rounds.DummyTxPayload"
        expected = {
            f"agent_{i}": {
                "_metaclass_registry_key": mcs_key,
                "id_": self.tx_payloads[i].id_,
                "round_count": self.tx_payloads[i].round_count,
                "sender": self.tx_payloads[i].sender,
                "value": self.tx_payloads[i].value,
                "vote": self.tx_payloads[i].vote,
            }
            for i in range(4)
        }

        assert self.test_round.serialized_collection == expected

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        round_id = DummyCollectionRound.auto_round_id()

        # collection round may set a flag to allow payments from inactive agents (rejoin)
        assert self.test_round._allow_rejoin_payloads is False  # default
        assert (
            self.test_round.accepting_payloads_from
            == self.synchronized_data.participants
        )
        self.test_round._allow_rejoin_payloads = True
        assert (
            self.test_round.accepting_payloads_from
            == self.synchronized_data.all_participants
        )

        first_payload, *_ = self.tx_payloads
        self.test_round.process_payload(first_payload)
        assert self.test_round.collection[first_payload.sender] == first_payload

        with pytest.raises(
            ABCIAppInternalError,
            match=f"internal error: sender agent_0 has already sent value for round: {round_id}",
        ):
            self.test_round.process_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape(
                "internal error: sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            self.test_round.process_payload(DummyTxPayload("sender", "value"))

        with pytest.raises(
            TransactionNotValidError,
            match=f"sender agent_0 has already sent value for round: {round_id}",
        ):
            self.test_round.check_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=re.escape(
                "sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            self.test_round.check_payload(DummyTxPayload("sender", "value"))

        self._test_payload_with_wrong_round_count(self.test_round)


class TestCollectDifferentUntilAllRound(_BaseRoundTestClass):
    """Test class for CollectDifferentUntilAllRound."""

    def test_run(
        self,
    ) -> None:
        """Run Tests."""

        test_round = DummyCollectDifferentUntilAllRound(
            synchronized_data=self.synchronized_data,
        )
        round_id = DummyCollectDifferentUntilAllRound.auto_round_id()

        first_payload, *payloads = self.tx_payloads
        test_round.process_payload(first_payload)
        assert not test_round.collection_threshold_reached

        with pytest.raises(
            ABCIAppInternalError,
            match=f"internal error: sender agent_0 has already sent value for round: {round_id}",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=f"sender agent_0 has already sent value for round: {round_id}",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: `CollectDifferentUntilAllRound` encountered a value '.*' that already exists.",
        ):
            object.__setattr__(first_payload, "sender", "other")
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="`CollectDifferentUntilAllRound` encountered a value '.*' that already exists.",
        ):
            test_round.check_payload(first_payload)

        for payload in payloads:
            assert not test_round.collection_threshold_reached
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached
        self._test_payload_with_wrong_round_count(test_round)


class TestCollectSameUntilAllRound(_BaseRoundTestClass):
    """Test class for CollectSameUntilAllRound."""

    def test_run(
        self,
    ) -> None:
        """Run Tests."""

        test_round = DummyCollectSameUntilAllRound(
            synchronized_data=self.synchronized_data,
        )
        round_id = DummyCollectSameUntilAllRound.auto_round_id()

        first_payload, *payloads = [
            DummyTxPayload(
                sender=agent,
                value="test",
            )
            for agent in sorted(self.participants)
        ]
        test_round.process_payload(first_payload)
        assert not test_round.collection_threshold_reached

        with pytest.raises(
            ABCIAppInternalError,
            match="1 votes are not enough for `CollectSameUntilAllRound`",
        ):
            assert test_round.common_payload

        with pytest.raises(
            ABCIAppInternalError,
            match=f"internal error: sender agent_0 has already sent value for round: {round_id}",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=f"sender agent_0 has already sent value for round: {round_id}",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: `CollectSameUntilAllRound` encountered a value '.*' "
            "which is not the same as the already existing one: '.*'",
        ):
            bad_payload = DummyTxPayload(
                sender="other",
                value="other",
            )
            test_round.process_payload(bad_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="`CollectSameUntilAllRound` encountered a value '.*' "
            "which is not the same as the already existing one: '.*'",
        ):
            test_round.check_payload(bad_payload)

        for payload in payloads:
            assert not test_round.collection_threshold_reached
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached
        assert test_round.common_payload
        self._test_payload_with_wrong_round_count(test_round, "test")


class TestCollectSameUntilThresholdRound(_BaseRoundTestClass):
    """Test CollectSameUntilThresholdRound."""

    @pytest.mark.parametrize(
        "selection_key",
        ("dummy_selection_key", tuple(f"dummy_selection_key_{i}" for i in range(2))),
    )
    def test_run(
        self,
        selection_key: Union[str, Tuple[str, ...]],
    ) -> None:
        """Run tests."""

        test_round = DummyCollectSameUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        test_round.collection_key = "dummy_collection_key"
        test_round.selection_key = selection_key
        assert test_round.end_block() is None

        first_payload, *payloads = get_dummy_tx_payloads(
            self.participants, value="vote"
        )
        test_round.process_payload(first_payload)

        assert not test_round.threshold_reached
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.threshold_reached
        assert test_round.most_voted_payload == "vote"

        self._test_payload_with_wrong_round_count(test_round)

        test_round.done_event = DummyEvent.DONE
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == test_round.done_event

        test_round.none_event = DummyEvent.NONE
        test_round.collection.clear()
        payloads = get_dummy_tx_payloads(
            self.participants, value=None, is_value_none=True, is_vote_none=True
        )
        for payload in payloads:
            test_round.process_payload(payload)
        assert test_round.most_voted_payload is None
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == test_round.none_event

        test_round.no_majority_event = DummyEvent.NO_MAJORITY
        test_round.collection.clear()
        for participant in self.participants:
            payload = DummyTxPayload(participant, value=participant)
            test_round.process_payload(payload)
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == test_round.no_majority_event

    def test_run_with_none(
        self,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectSameUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )

        first_payload, *payloads = get_dummy_tx_payloads(
            self.participants,
            value=None,
            is_value_none=True,
        )
        test_round.process_payload(first_payload)

        assert not test_round.threshold_reached
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.threshold_reached
        assert test_round.most_voted_payload is None


class TestOnlyKeeperSendsRound(_BaseRoundTestClass, BaseOnlyKeeperSendsRoundTest):
    """Test OnlyKeeperSendsRound."""

    @pytest.mark.parametrize(
        "payload_key", ("dummy_key", tuple(f"dummy_key_{i}" for i in range(2)))
    )
    def test_run(
        self,
        payload_key: Union[str, Tuple[str, ...]],
    ) -> None:
        """Run tests."""

        test_round = DummyOnlyKeeperSendsRound(
            synchronized_data=self.synchronized_data.update(
                most_voted_keeper_address="agent_0"
            ),
        )

        assert test_round.keeper_payload is None
        first_payload, *_ = self.tx_payloads
        test_round.process_payload(first_payload)
        assert test_round.keeper_payload is not None

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: keeper already set the payload.",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape(
                "internal error: sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.process_payload(DummyTxPayload(sender="sender", value="sender"))

        with pytest.raises(
            ABCIAppInternalError, match="internal error: agent_1 not elected as keeper."
        ):
            test_round.process_payload(DummyTxPayload(sender="agent_1", value="sender"))

        with pytest.raises(
            TransactionNotValidError, match="keeper payload value already set."
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=re.escape(
                "sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.check_payload(DummyTxPayload(sender="sender", value="sender"))

        with pytest.raises(
            TransactionNotValidError, match="agent_1 not elected as keeper."
        ):
            test_round.check_payload(DummyTxPayload(sender="agent_1", value="sender"))

        self._test_payload_with_wrong_round_count(test_round)

        test_round.done_event = DummyEvent.DONE
        test_round.payload_key = payload_key
        assert test_round.end_block()

    def test_keeper_payload_is_none(
        self,
    ) -> None:
        """Test keeper payload valur set to none."""

        keeper = "agent_0"
        self._complete_run(
            self._test_round(
                test_round=DummyOnlyKeeperSendsRound(
                    synchronized_data=self.synchronized_data.update(
                        most_voted_keeper_address=keeper,
                    ),
                ),
                keeper_payloads=DummyTxPayload(keeper, None),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data,
                synchronized_data_attr_checks=[],
                exit_event="FAIL_EVENT",
            )
        )


class TestVotingRound(_BaseRoundTestClass):
    """Test VotingRound."""

    def setup_test_voting_round(self) -> DummyVotingRound:
        """Setup test voting round"""
        return DummyVotingRound(
            synchronized_data=self.synchronized_data,
        )

    def test_vote_count(self) -> None:
        """Testing agent vote count"""
        test_round = self.setup_test_voting_round()
        a, b, c, d = self.participants
        for agents, vote in [((a, d), True), ((c,), False), ((b,), None)]:
            for payload in get_dummy_tx_payloads(frozenset(agents), vote=vote):
                test_round.process_payload(payload)
        assert dict(test_round.vote_count) == {True: 2, False: 1, None: 1}

        self._test_payload_with_wrong_round_count(test_round)

    @pytest.mark.parametrize("vote", [True, False, None])
    def test_threshold(self, vote: Optional[bool]) -> None:
        """Runs threshold test."""

        test_round = self.setup_test_voting_round()
        test_round.collection_key = "dummy_collection_key"
        test_round.done_event = DummyEvent.DONE
        test_round.negative_event = DummyEvent.NEGATIVE
        test_round.none_event = DummyEvent.NONE

        expected_threshold = {
            True: lambda: test_round.positive_vote_threshold_reached,
            False: lambda: test_round.negative_vote_threshold_reached,
            None: lambda: test_round.none_vote_threshold_reached,
        }[vote]

        expected_event = {
            True: test_round.done_event,
            False: test_round.negative_event,
            None: test_round.none_event,
        }[vote]

        first_payload, *payloads = get_dummy_tx_payloads(self.participants, vote=vote)
        test_round.process_payload(first_payload)
        assert test_round.end_block() is None
        assert not expected_threshold()
        for payload in payloads:
            test_round.process_payload(payload)
        assert expected_threshold()
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == expected_event

    def test_end_round_no_majority(self) -> None:
        """Test end round"""

        test_round = self.setup_test_voting_round()
        test_round.no_majority_event = DummyEvent.NO_MAJORITY
        for i, participant in enumerate(self.participants):
            payload = DummyTxPayload(participant, value=participant, vote=bool(i % 2))
            test_round.process_payload(payload)
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == test_round.no_majority_event

    def test_invalid_vote_payload_count(self) -> None:
        """Testing agent vote count with invalid payload."""
        test_round = self.setup_test_voting_round()
        a, b, c, d = self.participants

        class InvalidPayload(BaseTxPayload):
            """InvalidPayload"""

        def get_dummy_tx_payloads_(
            participants: FrozenSet[str],
        ) -> List[BaseTxPayload]:
            """Returns a list of DummyTxPayload objects."""
            return [InvalidPayload(sender=agent) for agent in sorted(participants)]

        for agents in [(a, d), (c,), (b,)]:
            for payload in get_dummy_tx_payloads_(frozenset(agents)):
                test_round.process_payload(payload)

        with pytest.raises(ValueError):
            test_round.vote_count


class TestCollectDifferentUntilThresholdRound(_BaseRoundTestClass):
    """Test CollectDifferentUntilThresholdRound."""

    @pytest.mark.parametrize(
        "required_confirmations", (MAX_PARTICIPANTS, MAX_PARTICIPANTS + 1)
    )
    def test_run(
        self,
        required_confirmations: int,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectDifferentUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        test_round.block_confirmations = 0
        test_round.required_block_confirmations = required_confirmations
        test_round.collection_key = "collection_key"
        test_round.done_event = 0
        assert (
            test_round.synchronized_data.consensus_threshold <= required_confirmations
        ), "Incorrect test parametrization: required confirmations cannot be set with a smalled value than the consensus threshold"

        first_payload, *payloads = get_dummy_tx_payloads(self.participants, vote=False)
        test_round.process_payload(first_payload)

        assert not test_round.collection_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)
            res = test_round.end_block()
            assert test_round.block_confirmations <= required_confirmations
            assert res is None
        assert test_round.collection_threshold_reached
        payloads_since_consensus = 2
        confirmations_remaining = required_confirmations - payloads_since_consensus
        for _ in range(confirmations_remaining):
            res = test_round.end_block()
            assert test_round.block_confirmations <= required_confirmations
            assert res is None

        res = test_round.end_block()
        assert test_round.block_confirmations > required_confirmations
        assert res is not None
        assert res[1] == test_round.done_event

        assert test_round.collection_threshold_reached
        self._test_payload_with_wrong_round_count(test_round)

    def test_end_round(self) -> None:
        """Test end round"""

        test_round = DummyCollectDifferentUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        test_round.collection_key = "dummy_collection_key"
        test_round.done_event = DummyEvent.DONE

        assert test_round.end_block() is None
        for participant in self.participants:
            payload = DummyTxPayload(participant, value=participant)
            test_round.process_payload(payload)
        return_value = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert return_value[-1] == test_round.done_event


class TestCollectNonEmptyUntilThresholdRound(_BaseRoundTestClass):
    """Test `CollectNonEmptyUntilThresholdRound`."""

    def test_get_non_empty_values(self) -> None:
        """Test `_get_non_empty_values`."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        payloads = get_dummy_tx_payloads(self.participants)
        none_payload_idx = 3
        object.__setattr__(payloads[none_payload_idx], "value", None)
        for payload in payloads:
            test_round.process_payload(payload)

        non_empty_values = test_round._get_non_empty_values()
        assert non_empty_values == {
            tuple(sorted(self.participants))[i]: (f"agent_{i}", False)
            if i != none_payload_idx
            else (False,)
            for i in range(4)
        }

        self._test_payload_with_wrong_round_count(test_round)

    def test_process_payload(self) -> None:
        """Test `process_payload`."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        first_payload, *payloads = get_dummy_tx_payloads(self.participants)
        test_round.process_payload(first_payload)

        assert not test_round.collection_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached

    @pytest.mark.parametrize(
        "selection_key",
        ("dummy_selection_key", tuple(f"dummy_selection_key_{i}" for i in range(2))),
    )
    @pytest.mark.parametrize(
        "is_value_none, expected_event",
        ((True, DummyEvent.NONE), (False, DummyEvent.DONE)),
    )
    def test_end_block(
        self,
        selection_key: Union[str, Tuple[str, ...]],
        is_value_none: bool,
        expected_event: str,
    ) -> None:
        """Test `end_block` when collection threshold is reached."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
        )
        test_round.selection_key = selection_key
        payloads = get_dummy_tx_payloads(
            self.participants, is_value_none=is_value_none, is_vote_none=True
        )
        for payload in payloads:
            test_round.process_payload(payload)

        test_round.collection = {f"test_{i}": payloads[i] for i in range(len(payloads))}
        test_round.collection_key = "test"
        test_round.done_event = DummyEvent.DONE
        test_round.none_event = DummyEvent.NONE

        res = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert res[0].db == self.synchronized_data.db
        assert res[1] == expected_event



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the storing functionality of abstract round abci."""

# pylint: skip-file

import json
from pathlib import Path, PosixPath
from typing import Optional, cast

import pytest

from packages.valory.skills.abstract_round_abci.io_.store import (
    CustomStorerType,
    JSONStorer,
    Storer,
    SupportedFiletype,
    SupportedStorerType,
)


class TestStorer:
    """Tests for the `Storer`."""

    def setup(self) -> None:
        """Setup the tests."""
        self.path = "tmp"
        self.json_storer = Storer(SupportedFiletype.JSON, None, self.path)

    def __dummy_custom_storer(self) -> None:
        """A dummy custom storing function to use for the tests."""

    @staticmethod
    @pytest.mark.parametrize(
        "filetype, custom_storer, expected_storer",
        (
            (None, None, None),
            (SupportedFiletype.JSON, None, JSONStorer.serialize_object),
            (
                SupportedFiletype.JSON,
                __dummy_custom_storer,
                JSONStorer.serialize_object,
            ),
            (None, __dummy_custom_storer, __dummy_custom_storer),
        ),
    )
    def test__get_single_storer_from_filetype(
        filetype: Optional[SupportedFiletype],
        custom_storer: Optional[CustomStorerType],
        expected_storer: Optional[SupportedStorerType],
        tmp_path: PosixPath,
    ) -> None:
        """Test `_get_single_storer_from_filetype`."""
        if all(
            test_arg is None for test_arg in (filetype, custom_storer, expected_storer)
        ):
            with pytest.raises(
                ValueError,
                match="Please provide either a supported filetype or a custom storing function.",
            ):
                Storer(
                    filetype, custom_storer, str(tmp_path)
                )._get_single_storer_from_filetype()

        else:
            expected_storer = cast(SupportedStorerType, expected_storer)
            storer = Storer(filetype, custom_storer, str(tmp_path))
            assert (
                storer._get_single_storer_from_filetype().__code__.co_code
                == expected_storer.__code__.co_code
            )

    def test_store(self) -> None:
        """Test `store`."""
        dummy_object = {"test": "test"}
        expected_object = {self.path: json.dumps(dummy_object, indent=4)}
        actual_object = self.json_storer.store(dummy_object, False)
        assert expected_object == actual_object

    def test_store_multiple(self) -> None:
        """Test `store` when multiple files are present."""
        dummy_object = {"test": "test"}
        dummy_filename = "test"
        expected_path = Path(f"{self.path}/{dummy_filename}").__str__()
        expected_object = {expected_path: json.dumps(dummy_object, indent=4)}
        actual_object = self.json_storer.store({dummy_filename: dummy_object}, True)
        assert expected_object == actual_object



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test replay tools."""

import shutil
import subprocess  # nosec
import tempfile
from pathlib import Path
from typing import cast
from unittest import mock

import pytest

from autonomy.deploy.base import TENDERMINT_COM_URL_PARAM, TENDERMINT_URL_PARAM
from autonomy.deploy.constants import TM_STATE_DIR
from autonomy.replay.agent import AgentRunner
from autonomy.replay.tendermint import (
    RanOutOfDumpsToReplay,
    TendermintRunner,
    build_tendermint_apps,
)

from tests.conftest import ROOT_DIR


TENDERMINT_BIN = shutil.which("tendermint")
AGENT_DATA = {
    "mem_limit": "1024m",
    "mem_reservation": "256M",
    "cpus": 1,
    "container_name": "abci0",
    "image": "valory/open-autonomy-open-aea:hello-world-0.1.0",
    "environment": [
        "LOG_FILE=/logs/aea_0.txt",
        "ID=0",
        "VALORY_APPLICATION=valory/hello_world:0.1.0:bafybeideb6b5k4i6z7bm3p53eydxgknmwdefo2oshcnlxthjc6oxeox7ua",
        "ABCI_HOST=abci0",
        f"SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_{TENDERMINT_URL_PARAM.upper()}=http://node0:26657",
        f"SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_{TENDERMINT_COM_URL_PARAM.upper()}=http://node0:8080",
        "LEDGER_ADDRESS=http://143.110.184.220:8545",
        "LEDGER_CHAIN_ID=31337",
        "AEA_KEY=0x874741e86698d72ce5a579386ab6f3e006426e4959662736aa3132b83911130a",
    ],
    "networks": {"localnet": {"ipv4_address": "192.167.11.7"}},
    "volumes": ["./persistent_data/logs:/logs:Z", "./agent_keys/agent_0:/agent_key:Z"],
}


def ctrl_c() -> None:
    """Generate keyboard inturrupt."""
    raise KeyboardInterrupt()


def init_tendermint(home: Path) -> None:
    """Initialize tendermint home."""
    result = subprocess.run(  # nosec
        [cast(str, TENDERMINT_BIN), "init", "--home", str(home)]
    )
    assert result.returncode == 0, result.stdout


def test_tendermint_runner() -> None:
    """Test `TendermintRunner` class."""

    number_of_periods = 2
    node_id = 0

    with tempfile.TemporaryDirectory() as temp_dir:
        dump_dir = Path(temp_dir, TM_STATE_DIR)
        dump_dir.mkdir()

        for p in range(number_of_periods):
            tm_home = dump_dir / f"period_{p}" / f"node{node_id}"
            init_tendermint(tm_home)

        runner = TendermintRunner(
            node_id=node_id,
            dump_dir=dump_dir,
            n_periods=number_of_periods,
        )

        assert runner.get_last_block_height() == 0

        with pytest.raises(ValueError, match="Cannot find tendermint installation."):
            with mock.patch("shutil.which", new=lambda _: None):
                runner.start()

        runner.start()
        assert isinstance(runner.process, subprocess.Popen)

        runner.stop()
        assert runner.process is None

        runner.update_period()
        assert runner.period == 1

        with pytest.raises(RanOutOfDumpsToReplay):
            runner.update_period()


def test_tendermint_network() -> None:
    """Test tendermint network."""

    periods = 2
    nodes = 2
    app, tendermint_network = build_tendermint_apps()

    with tempfile.TemporaryDirectory() as temp_dir:
        dump_dir = Path(temp_dir, TM_STATE_DIR)

        with pytest.raises(FileNotFoundError, match="Can't find period dumps in"):
            tendermint_network.init(dump_dir=dump_dir)

        dump_dir.mkdir()
        for p in range(periods):
            (dump_dir / f"period_{p}").mkdir()

        with pytest.raises(FileNotFoundError, match="Can't find dumped nodes"):
            tendermint_network.init(dump_dir=dump_dir)

        for p in range(periods):
            period_dir = dump_dir / f"period_{p}"
            for n in range(nodes):
                node_dir = period_dir / f"node{n}"
                init_tendermint(node_dir)

        tendermint_network.init(dump_dir=dump_dir)
        assert tendermint_network.get_last_block_height(0) == 0

        tendermint_network.start()
        for node in tendermint_network.nodes:
            assert isinstance(node.process, subprocess.Popen)

        tendermint_network.stop()
        for node in tendermint_network.nodes:
            assert node.process is None

        with mock.patch.object(tendermint_network, "start", ctrl_c):
            tendermint_network.run_until_interruption()

        with app.test_client() as client:
            response = client.get("/0/tx")
            response_data = response.get_json()
            assert response_data["result"]["tx_result"]["code"] == 0

            response = client.get("/0/broadcast_tx_sync")
            response_data = response.get_json()
            assert response_data["result"]["code"] == 0
            assert response_data["result"]["hash"] == ""

            response = client.get("/0/status")
            response_data = response.get_json()
            assert response_data["result"]["sync_info"]["latest_block_height"] == -1

            response = client.get("/0/hard_reset")
            response_data = response.get_json()
            assert response_data["status"] is True
            assert response_data["is_replay"] is True

            response = client.get("/0/hard_reset")
            response_data = response.get_json()
            assert response_data["status"] is False
            assert (
                response_data["message"]
                == "Ran out of dumps to replay, You can stop the agent replay now."
            )
            assert response_data["is_replay"] is True


def test_agent_runner() -> None:
    """Test agent runner."""

    agent_runner = AgentRunner(
        0,
        AGENT_DATA,
        ROOT_DIR / "packages",
    )

    agent_runner.start()
    assert isinstance(agent_runner.process, subprocess.Popen)
    assert (
        len(
            {
                "vendor",
                "ethereum_private_key.txt",
                "aea-config.yaml",
                "README.md",
            }.difference(
                set(
                    map(
                        lambda x: x.name,
                        Path(agent_runner.agent_dir.name, "agent").iterdir(),
                    )
                )
            )
        )
        == 0
    )

    agent_runner.stop()
    assert agent_runner.process is None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/gnosis contract."""

import binascii
import secrets
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast
from unittest import mock

import pytest
from aea.common import JSONLike
from aea.crypto.registries import crypto_registry
from aea_ledger_ethereum import EthereumApi, EthereumCrypto
from aea_test_autonomy.base_test_classes.contracts import (
    BaseGanacheContractTest,
    BaseHardhatGnosisContractTest,
)
from aea_test_autonomy.configurations import (
    ETHEREUM_KEY_PATH_1,
    ETHEREUM_KEY_PATH_2,
    ETHEREUM_KEY_PATH_3,
    KEY_PAIRS,
)
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.helpers.contracts import get_register_contract
from hexbytes import HexBytes
from web3 import Web3
from web3.datastructures import AttributeDict
from web3.eth import Eth
from web3.exceptions import SolidityError
from web3.types import TxData

from packages.valory.contracts.gnosis_safe.contract import (
    GnosisSafeContract,
    SAFE_CONTRACT,
)
from packages.valory.contracts.gnosis_safe_proxy_factory.tests.test_contract import (
    PACKAGE_DIR as PROXY_DIR,
)


PACKAGE_DIR = Path(__file__).parent.parent

DEFAULT_GAS = 1000000
DEFAULT_MAX_FEE_PER_GAS = 10**15
DEFAULT_MAX_PRIORITY_FEE_PER_GAS = 10**15
EXPECTED_TX_KEYS = {
    "chainId",
    "data",
    "from",
    "gas",
    "maxFeePerGas",
    "maxPriorityFeePerGas",
    "nonce",
    "to",
    "value",
}


class BaseContractTest(BaseGanacheContractTest):
    """Base test case for GnosisSafeContract"""

    NB_OWNERS: int = 4
    THRESHOLD: int = 1
    SALT_NONCE: Optional[int] = None
    contract_directory = PACKAGE_DIR
    contract: GnosisSafeContract

    @classmethod
    def setup_class(
        cls,
    ) -> None:
        """Setup test."""
        # workaround for the fact that contract dependencies are not possible yet
        _ = get_register_contract(PROXY_DIR)
        super().setup_class()

    @classmethod
    def deployment_kwargs(cls) -> Dict[str, Any]:
        """Get deployment kwargs."""
        return dict(
            owners=cls.owners(),
            threshold=int(cls.threshold()),
            gas=DEFAULT_GAS,
        )

    @classmethod
    def owners(cls) -> List[str]:
        """Get the owners."""
        return [Web3.toChecksumAddress(t[0]) for t in cls.key_pairs()[: cls.NB_OWNERS]]

    @classmethod
    def deployer(cls) -> Tuple[str, str]:
        """Get the key pair of the deployer."""
        # for simplicity, get the first owner
        return cls.key_pairs()[0]

    @classmethod
    def threshold(
        cls,
    ) -> int:
        """Returns the amount of threshold."""
        return cls.THRESHOLD

    @classmethod
    def get_nonce(cls) -> int:
        """Get the nonce."""
        if cls.SALT_NONCE is not None:
            return cls.SALT_NONCE
        return secrets.SystemRandom().randint(0, 2**256 - 1)


class BaseContractTestHardHatSafeNet(BaseHardhatGnosisContractTest):
    """Base test case for GnosisSafeContract"""

    NB_OWNERS: int = 4
    THRESHOLD: int = 1
    SALT_NONCE: Optional[int] = None
    contract_directory = PACKAGE_DIR
    sanitize_from_deploy_tx = ["contract_address"]
    contract: GnosisSafeContract

    @classmethod
    def setup_class(
        cls,
    ) -> None:
        """Setup test."""
        _ = get_register_contract(PROXY_DIR)
        super().setup_class()

    @classmethod
    def deployment_kwargs(cls) -> Dict[str, Any]:
        """Get deployment kwargs."""
        return dict(
            owners=cls.owners(),
            threshold=int(cls.threshold()),
            gas=DEFAULT_GAS,
        )

    @classmethod
    def owners(cls) -> List[str]:
        """Get the owners."""
        return [Web3.toChecksumAddress(t[0]) for t in cls.key_pairs()[: cls.NB_OWNERS]]

    @classmethod
    def deployer(cls) -> Tuple[str, str]:
        """Get the key pair of the deployer."""
        # for simplicity, get the first owner
        return cls.key_pairs()[0]

    @classmethod
    def threshold(
        cls,
    ) -> int:
        """Returns the amount of threshold."""
        return cls.THRESHOLD

    @classmethod
    def get_nonce(cls) -> int:
        """Get the nonce."""
        if cls.SALT_NONCE is not None:
            return cls.SALT_NONCE
        return secrets.SystemRandom().randint(0, 2**256 - 1)

    def _verify_safe_tx(
        self,
        tx_hash: str,
        value: int,
        data: bytes,
        to_address: str,
        signatures_by_owners: Dict,
    ) -> bool:
        """Helper to verify tx."""
        ledger_api = cast(EthereumApi, self.ledger_api)
        contract_address = cast(str, self.contract_address)
        return self.contract.verify_tx(
            ledger_api=ledger_api,
            contract_address=contract_address,
            tx_hash=tx_hash,
            owners=(self.deployer_crypto.address.lower(),),
            to_address=to_address,
            value=value,
            data=data,
            signatures_by_owner={
                self.deployer_crypto.address.lower(): signatures_by_owners[
                    self.deployer_crypto.address
                ]
            },
        )["verified"]

    def _get_raw_safe_tx(
        self,
        sender_address: str,
        value: int,
        data: bytes,
        to_address: str,
        signatures_by_owners: Dict,
    ) -> JSONLike:
        """Helper to prepare a safe tx."""
        ledger_api = cast(EthereumApi, self.ledger_api)
        contract_address = cast(str, self.contract_address)
        tx = self.contract.get_raw_safe_transaction(
            ledger_api=ledger_api,
            contract_address=contract_address,
            sender_address=sender_address,
            owners=(self.deployer_crypto.address.lower(),),
            to_address=to_address,
            value=value,
            data=data,
            signatures_by_owner={
                self.deployer_crypto.address.lower(): signatures_by_owners[
                    self.deployer_crypto.address
                ]
            },
        )
        return tx


@skip_docker_tests
class TestDeployTransactionHardhat(BaseContractTestHardHatSafeNet):
    """Test."""

    ledger_api: EthereumApi

    def test_deployed(self) -> None:
        """Run tests."""
        result = self.contract.get_deploy_transaction(
            ledger_api=self.ledger_api,
            deployer_address=str(self.deployer_crypto.address),
            owners=self.owners(),
            threshold=int(self.threshold()),
            gas=DEFAULT_GAS,
        )
        assert isinstance(result, dict)
        assert len(result) == 10
        data = result.pop("data")
        assert isinstance(data, str)
        assert len(data) > 0 and data.startswith("0x")
        # there is no "data" field on the deploy tx
        # but there is contract_address
        expected_deploy_tx_fields = EXPECTED_TX_KEYS - {"data"}
        expected_deploy_tx_fields.add("contract_address")
        assert all(
            key in result.keys() for key in expected_deploy_tx_fields
        ), "Error, found: {}".format(result)

    def test_exceptions(
        self,
    ) -> None:
        """Test exceptions."""

        with pytest.raises(
            ValueError,
            match="Threshold cannot be bigger than the number of unique owners",
        ):
            # Tests for `ValueError("Threshold cannot be bigger than the number of unique owners")`.`
            self.contract.get_deploy_transaction(
                ledger_api=self.ledger_api,
                deployer_address=str(self.deployer_crypto.address),
                owners=[],
                threshold=1,
            )

        with pytest.raises(ValueError, match="Client does not have any funds"):
            # Tests for `ValueError("Client does not have any funds")`.
            self.contract.get_deploy_transaction(
                ledger_api=self.ledger_api,
                deployer_address=crypto_registry.make(
                    EthereumCrypto.identifier
                ).address,
                owners=self.owners(),
                threshold=int(self.threshold()),
            )

    def test_non_implemented_methods(
        self,
    ) -> None:
        """Test not implemented methods."""
        with pytest.raises(NotImplementedError):
            self.contract.get_raw_transaction(self.ledger_api, "")

        with pytest.raises(NotImplementedError):
            self.contract.get_raw_message(self.ledger_api, "")

        with pytest.raises(NotImplementedError):
            self.contract.get_state(self.ledger_api, "")

    def test_verify(self) -> None:
        """Run verify test."""
        assert self.contract_address is not None
        result = self.contract.verify_contract(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
        )
        assert result["verified"], "Contract not verified."

        verified = self.contract.verify_contract(
            ledger_api=self.ledger_api,
            contract_address=SAFE_CONTRACT,
        )["verified"]
        assert not verified, "Not verified"

    def test_get_safe_nonce(self) -> None:
        """Run get_safe_nonce test."""
        safe_nonce = self.contract.get_safe_nonce(
            ledger_api=self.ledger_api,
            contract_address=cast(str, self.contract_address),
        )["safe_nonce"]
        assert safe_nonce == 0

    def test_revert_reason(
        self,
    ) -> None:
        """Test `revert_reason` method."""
        tx = {
            "to": "to",
            "from": "from",
            "value": "value",
            "input": "input",
            "blockNumber": 1,
        }

        def _raise_solidity_error(*_: Any) -> None:
            raise SolidityError("reason")

        with mock.patch.object(
            self.ledger_api.api.eth, "call", new=_raise_solidity_error
        ):
            reason = self.contract.revert_reason(
                self.ledger_api, "contract_address", cast(TxData, tx)
            )
            assert "revert_reason" in reason
            assert reason["revert_reason"] == "SolidityError('reason')"

        with mock.patch.object(self.ledger_api.api.eth, "call"), pytest.raises(
            ValueError, match=f"The given transaction has not been reverted!\ntx: {tx}"
        ):
            self.contract.revert_reason(
                self.ledger_api, "contract_address", cast(TxData, tx)
            )

    def test_get_incoming_transfers(self) -> None:
        """Run get_incoming txs."""
        from_block = cast(int, self.ledger_api.api.eth.get_block_number()) + 1
        self.ledger_api.api.eth.send_transaction(
            {
                "to": self.contract_address,
                "from": self.deployer_crypto.address,
                "value": 10,
            }
        )

        res = cast(
            JSONLike,
            self.contract.get_ingoing_transfers(
                ledger_api=self.ledger_api,
                contract_address=cast(str, self.contract_address),
                from_block=hex(from_block),
            ),
        )
        data = cast(List[JSONLike], res["data"])

        time.sleep(1)

        assert len(res) == 1, "one transfer should exist"
        assert data[0]["amount"] == 10, "transfer amount should be 10"
        assert (
            data[0]["sender"] == self.deployer_crypto.address
        ), f"{data[0]['sender']} should be the sender"
        assert data[0]["blockNumber"] is not None, "tx is still pending"
        assert (
            self.ledger_api.api.eth.get_balance(self.contract_address) == 10
        ), "incorrect balance"

        self.ledger_api.api.eth.send_transaction(
            {
                "to": self.contract_address,
                "from": self.deployer_crypto.address,
                "value": 100,
            }
        )
        from_block = cast(int, data[0]["blockNumber"]) + 1

        time.sleep(3)
        res = self.contract.get_ingoing_transfers(
            ledger_api=self.ledger_api,
            contract_address=cast(str, self.contract_address),
            from_block=hex(from_block),
        )
        data = cast(List[JSONLike], res["data"])

        assert len(res) == 1, "one transfer should exist"
        assert data[0]["amount"] == 100, "transfer amount should be 100"
        assert (
            data[0]["sender"] == self.deployer_crypto.address
        ), f"{data[0]['sender']} should be the sender"
        assert data[0]["blockNumber"] is not None, "tx is still pending"
        assert (
            self.ledger_api.api.eth.get_balance(self.contract_address) == 110
        ), "incorrect balance"

    def test_get_zero_transfer_events(self) -> None:
        """Test get_zero_transfer_events."""
        # check that the safe has no zero transfers
        res = cast(
            JSONLike,
            self.contract.get_zero_transfer_events(
                ledger_api=self.ledger_api,
                contract_address=cast(str, self.contract_address),
                sender_address=self.deployer_crypto.address,
            ),
        )
        data = cast(List[JSONLike], res["data"])
        assert len(data) == 0, "no zero transfers should be in the safe"

        # make a zero transfer
        self.ledger_api.api.eth.send_transaction(
            {
                "to": self.contract_address,
                "from": self.deployer_crypto.address,
                "value": 0,
            }
        )

        time.sleep(3)

        res = cast(
            JSONLike,
            self.contract.get_zero_transfer_events(
                ledger_api=self.ledger_api,
                contract_address=cast(str, self.contract_address),
                sender_address=self.deployer_crypto.address,
            ),
        )
        data = cast(List[JSONLike], res["data"])

        assert len(res) == 1, "one zero transfer should exist"
        assert (
            data[0]["sender"] == self.deployer_crypto.address
        ), f"{data[0]['sender']} should be the sender"
        assert data[0]["block_number"] is not None, "tx is still pending"

        # make a second zero transfer
        prev_block = cast(int, self.ledger_api.api.eth.get_block_number()) + 1
        self.ledger_api.api.eth.send_transaction(
            {
                "to": self.contract_address,
                "from": self.deployer_crypto.address,
                "value": 0,
            }
        )
        time.sleep(3)

        res = self.contract.get_zero_transfer_events(
            ledger_api=self.ledger_api,
            contract_address=cast(str, self.contract_address),
            sender_address=self.deployer_crypto.address,
            from_block=prev_block,
        )
        data = cast(List[JSONLike], res["data"])

        assert len(res) == 1, "one zero transfer should exist"
        assert (
            data[0]["sender"] == self.deployer_crypto.address
        ), f"{data[0]['sender']} should be the sender"
        assert data[0]["block_number"] is not None, "tx is still pending"

    def test_get_owners(self) -> None:
        """Test the owners are as expected."""
        owners = self.contract.get_owners(
            ledger_api=self.ledger_api,
            contract_address=cast(str, self.contract_address),
        )["owners"]
        assert owners == self.owners()


@skip_docker_tests
class TestRawSafeTransaction(BaseContractTestHardHatSafeNet):
    """Test `get_raw_safe_transaction`"""

    ledger_api: EthereumApi

    def test_run(self) -> None:
        """Run tests."""
        assert self.contract_address is not None
        value = 0
        data = b""
        sender = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_1
        )
        assert sender.address == self.owners()[1]
        receiver = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_2
        )
        assert receiver.address == self.owners()[2]
        fourth = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_3
        )
        assert fourth.address == self.owners()[3]
        cryptos = [self.deployer_crypto, sender, receiver, fourth]
        tx_hash = self.contract.get_raw_safe_transaction_hash(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            to_address=receiver.address,
            value=value,
            data=data,
        )["tx_hash"]
        b_tx_hash = binascii.unhexlify(cast(str, tx_hash)[2:])
        signatures_by_owners = {
            crypto.address: crypto.sign_message(b_tx_hash, is_deprecated_mode=True)[2:]
            for crypto in cryptos
        }
        assert list(signatures_by_owners.keys()) == self.owners()

        tx = self._get_raw_safe_tx(
            sender_address=sender.address,
            value=value,
            data=data,
            signatures_by_owners=signatures_by_owners,
            to_address=receiver.address,
        )
        assert all(key in tx.keys() for key in EXPECTED_TX_KEYS), "Missing key"

        tx_signed = sender.sign_transaction(tx)
        tx_hash = self.ledger_api.send_signed_transaction(tx_signed)
        assert tx_hash is not None, "Tx hash is `None`"

        verified = self._verify_safe_tx(
            tx_hash, value, data, receiver.address, signatures_by_owners
        )
        assert verified, f"Not verified: {verified}"

    def test_verify_negative(self) -> None:
        """Test verify negative."""
        assert self.contract_address is not None
        verified = self.contract.verify_tx(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            tx_hash="0xfc6d7c491688840e79ed7d8f0fc73494be305250f0d5f62d04c41bc4467e8603",
            owners=("",),
            to_address=crypto_registry.make(
                EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_1
            ).address,
            value=0,
            data=b"",
            signatures_by_owner={},
        )["verified"]
        assert not verified, "Should not be verified"

    # mock `get_transaction` and `get_transaction_receipt` using a copy of a real reverted tx and its receipt
    @mock.patch.object(
        Eth,
        "get_transaction",
        return_value=AttributeDict(
            {
                "accessList": [],
                "blockHash": HexBytes(
                    "0x8543592f08d1d9e6d722ba9d600270d7e7789ecc9b66f27ca81b104df9c5dd4a"
                ),
                "blockNumber": 31190129,
                "chainId": "0x89",
                "from": "0x5eF6567079c6c26d8ebf61AC0716163367E9B3cf",
                "gas": 270000,
                "gasPrice": 36215860217,
                "hash": HexBytes(
                    "0x09d5be525caea564b2d4fd31af424c8f0414a9b270937a1bee29167a883e6ce5"
                ),
                "input": "0x6a7612020000000000000000000000003d9e92b0fe7673dda3d7c33b9ff302768a03de190000000000000000000"
                "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
                "000000000000014000000000000000000000000000000000000000000000000000000000000000000000000000000"
                "00000000000000000000000000000000000000000000001d4c0000000000000000000000000000000000000000000"
                "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
                "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
                "000000000000000000000000000000000000000000000000000000000000000000000000000000000000000002000"
                "0000000000000000000000000000000000000000000000000000000000000846b0bac970000000000000000000000"
                "000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000"
                "000000000004000000000000000000000000000000000000000000000000000000000001df5000000000000000000"
                "00000000000000000000000000000487da3583c85e1e0000000000000000000000000000000000000000000000000"
                "000000000000000000000000000000000000000000000000000000000000000000000000104c292f99a14d354c669"
                "3f9037a4a3d09c85c8ad5f1ab4de79bbc8bab845560f797f385ecbe77e90245b7b45e218a2c56fec17c9d38729264"
                "83d0ed800df46daa71c3afaa87b5959d644cd0d311a93acb398ec4f9d4c545c54ea6f4adbaa3e99dd9668f948eb64"
                "10f1b2105e2f6ca762badf17539d9221cef7af55a244c6ae3c6b401cfd01fe829d711a372b9d8ad5b91e0956a4da1"
                "6929d04a2581b10f9f4599899b625c367bef18656c90efcf9d9ee5063860774f08517488b05ef5090acd31aa9d91b"
                "7df8080d69fdddfe9b326f3ae0cb95227e21d2d265b6a83861998dd9e91fb980415e78c2bb0b10dbe3b4d7bead977"
                "2f32fa26b738c5670aa69ee9d09973ea2b81c00000000000000000000000000000000000000000000000000000000",
                "maxFeePerGas": 36215860217,
                "maxPriorityFeePerGas": 36215860202,
                "nonce": 2231,
                "r": HexBytes(
                    "0x5d5d369d5fc30c5604d974761d41b08118120eb94fd65a827bab1f6ea558d67c"
                ),
                "s": HexBytes(
                    "0x12f68826bd41989335e62d43fd36547fe171ad536b99bc93766622438d3f8355"
                ),
                "to": "0x37ba5291A5bE8cbE44717a0673fe2c5a45B4B6A8",
                "transactionIndex": 28,
                "type": "0x2",
                "v": 1,
                "value": 0,
            }
        ),
    )
    @mock.patch.object(
        EthereumApi,
        "get_transaction_receipt",
        return_value={
            "blockHash": "0x8543592f08d1d9e6d722ba9d600270d7e7789ecc9b66f27ca81b104df9c5dd4a",
            "blockNumber": 31190129,
            "contractAddress": None,
            "cumulativeGasUsed": 5167853,
            "effectiveGasPrice": 36215860217,
            "from": "0x5eF6567079c6c26d8ebf61AC0716163367E9B3cf",
            "gasUsed": 48921,
            "logs": [
                {
                    "address": "0x0000000000000000000000000000000000001010",
                    "blockHash": "0x8543592f08d1d9e6d722ba9d600270d7e7789ecc9b66f27ca81b104df9c5dd4a",
                    "blockNumber": 31190129,
                    "data": "0x00000000000000000000000000000000000000000000000000064b5dcc9920c1000000000000000000000000"
                    "00000000000000000000000032116d529b00f7490000000000000000000000000000000000000000000004353d"
                    "1a5e0a73394e1e000000000000000000000000000000000000000000000000320b21f4ce67d688000000000000"
                    "0000000000000000000000000000000004353d20a9683fd26edf",
                    "logIndex": 115,
                    "removed": False,
                    "topics": [
                        "0x4dfe1bbbcf077ddc3e01291eea2d5c70c2b422b415d95645b9adcfd678cb1d63",
                        "0x0000000000000000000000000000000000000000000000000000000000001010",
                        "0x0000000000000000000000005ef6567079c6c26d8ebf61ac0716163367e9b3cf",
                        "0x000000000000000000000000f0245f6251bef9447a08766b9da2b07b28ad80b0",
                    ],
                    "transactionHash": "0x09d5be525caea564b2d4fd31af424c8f0414a9b270937a1bee29167a883e6ce5",
                    "transactionIndex": 28,
                }
            ],
            "logsBloom": "0x0000000000000000000000000000000000000000000000000000000000000000000000000000800000000000000"
            "000000000800000000000000000000000000000008000000000000000000000000080000000000000000000010000"
            "000000000000000000000000000000000000000000000000000000008000000000000000000000008000000000000"
            "000000000000000000000000000000000000088000020000000000000000000000000000000000000000000000000"
            "000000000000400000000000000000000100000000000000000000000000000010000000000000000000000000000"
            "0000000800000000000000000000000000000000000100000",
            "status": 0,
            "to": "0x37ba5291A5bE8cbE44717a0673fe2c5a45B4B6A8",
            "transactionHash": "0x09d5be525caea564b2d4fd31af424c8f0414a9b270937a1bee29167a883e6ce5",
            "transactionIndex": 28,
            "type": "0x2",
            "revert_reason": "execution reverted: GS026",
        },
    )
    def test_verify_reverted(self, *_: Any) -> None:
        """Test verify for reverted tx."""
        assert self.contract_address is not None
        res = self.contract.verify_tx(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            tx_hash="test",
            owners=("",),
            to_address=crypto_registry.make(
                EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_1
            ).address,
            value=0,
            data=b"",
            signatures_by_owner={},
        )
        assert not res["verified"], "Should not be verified"


@skip_docker_tests
class TestOwnerManagement(BaseContractTestHardHatSafeNet):
    """Test owner management related ."""

    ledger_api: EthereumApi

    def test_remove(self) -> None:  # pylint: disable=too-many-locals
        """Test owner removal."""
        assert self.contract_address is not None
        owner_to_be_removed = self.owners()[2]
        threshold = max(self.threshold() - 1, 1)
        value = 0
        data_str = self.contract.get_remove_owner_data(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            owner=owner_to_be_removed,
            threshold=threshold,
        ).get("data")
        data = bytes.fromhex(data_str[2:])  # strip 0x before converting to bytes

        sender = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_1
        )
        assert sender.address == self.owners()[1]
        receiver = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_2
        )
        assert receiver.address == self.owners()[2]
        fourth = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_3
        )
        assert fourth.address == self.owners()[3]
        cryptos = [self.deployer_crypto, sender, receiver, fourth]
        tx_hash = self.contract.get_raw_safe_transaction_hash(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            to_address=self.contract_address,
            value=value,
            data=data,
        )["tx_hash"]
        b_tx_hash = binascii.unhexlify(cast(str, tx_hash)[2:])
        signatures_by_owners = {
            crypto.address: crypto.sign_message(b_tx_hash, is_deprecated_mode=True)[2:]
            for crypto in cryptos
        }
        assert list(signatures_by_owners.keys()) == self.owners()

        tx = self._get_raw_safe_tx(
            sender_address=sender.address,
            value=value,
            data=data,
            to_address=self.contract_address,
            signatures_by_owners=signatures_by_owners,
        )
        assert all(key in tx.keys() for key in EXPECTED_TX_KEYS), "Missing key"

        prev_block = cast(int, self.ledger_api.api.eth.get_block_number()) + 1
        tx_signed = sender.sign_transaction(tx)
        tx_hash = self.ledger_api.send_signed_transaction(tx_signed)

        assert tx_hash is not None, "Tx hash is `None`"

        verified = self._verify_safe_tx(
            tx_hash, value, data, self.contract_address, signatures_by_owners
        )
        assert verified, f"Not verified: {verified}"
        time.sleep(1)
        remove_events = self.contract.get_removed_owner_events(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            removed_owner=owner_to_be_removed,
            from_block=prev_block,
        ).get("data")

        assert remove_events is not None, "a RemovedOwner event was expected"
        assert len(remove_events) == 1, "1 RemovedOwner event was expected"
        assert (
            remove_events[0].get("owner") == owner_to_be_removed
        ), "a different owner than expected was removed"

    def test_swap(self) -> None:  # pylint: disable=too-many-locals
        """Test owner swapping."""
        assert self.contract_address is not None
        old_owner = self.owners()[1]
        new_owner = KEY_PAIRS[-1][0]
        value = 0
        data_str = self.contract.get_swap_owner_data(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            old_owner=old_owner,
            new_owner=new_owner,
        ).get("data")
        data = bytes.fromhex(data_str[2:])  # strip 0x before converting to bytes

        sender = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_1
        )
        assert sender.address == self.owners()[1]
        receiver = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_2
        )
        assert receiver.address == self.owners()[2]
        fourth = crypto_registry.make(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_PATH_3
        )
        assert fourth.address == self.owners()[3]
        cryptos = [self.deployer_crypto, sender, receiver, fourth]
        tx_hash = self.contract.get_raw_safe_transaction_hash(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            to_address=self.contract_address,
            value=value,
            data=data,
        )["tx_hash"]
        b_tx_hash = binascii.unhexlify(cast(str, tx_hash)[2:])
        signatures_by_owners = {
            crypto.address: crypto.sign_message(b_tx_hash, is_deprecated_mode=True)[2:]
            for crypto in cryptos
        }
        assert list(signatures_by_owners.keys()) == self.owners()

        tx = self._get_raw_safe_tx(
            sender_address=sender.address,
            to_address=self.contract_address,
            value=value,
            data=data,
            signatures_by_owners=signatures_by_owners,
        )
        assert all(key in tx.keys() for key in EXPECTED_TX_KEYS), "Missing key"

        prev_block = cast(int, self.ledger_api.api.eth.get_block_number()) + 1
        tx_signed = sender.sign_transaction(tx)
        tx_hash = self.ledger_api.send_signed_transaction(tx_signed)

        assert tx_hash is not None, "Tx hash is `None`"

        verified = self._verify_safe_tx(
            tx_hash, value, data, self.contract_address, signatures_by_owners
        )
        assert verified, f"Not verified: {verified}"
        time.sleep(1)
        remove_events = self.contract.get_removed_owner_events(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            removed_owner=old_owner,
            from_block=prev_block,
        ).get("data")

        assert remove_events is not None, "a RemovedOwner event was expected"
        assert len(remove_events) == 1, "1 RemovedOwner event was expected"
        assert (
            remove_events[0].get("owner") == old_owner
        ), "a different owner than expected was removed"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test package manager API."""


import json
import tempfile
from pathlib import Path
from unittest import mock

from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.package_manager.base import DepedencyMismatchErrors
from aea.test_tools.test_cases import BaseAEATestCase

from autonomy.cli.packages import (
    PackageManagerV0,
    PackageManagerV1,
    get_package_manager,
)
from autonomy.configurations.base import Service
from autonomy.configurations.constants import DEFAULT_SERVICE_CONFIG_FILE


DUMMY_HASH = "bafybei0000000000000000000000000000000000000000000000000000"
SERVICE_ID = PackageId(
    package_type=PackageType.SERVICE,
    public_id=PublicId(author="valory", name="service"),
)
SERVICE_CONFIG = {
    "name": "dummy_service",
    "author": "valory",
    "version": "0.1.0",
    "description": "A dummy service config file.",
    "aea_version": ">=1.0.0, <2.0.0",
    "license": "Apache-2.0",
    "fingerprint": {
        "README.md": "bafybeiapubcoersqnsnh3acia5hd7otzt7kjxekr6gkbrlumv6tkajl6jm",
    },
    "fingerprint_ignore_patterns": "[]",
    "agent": f"valory/hello_world:0.1.0:{DUMMY_HASH}",
    "number_of_agents": "1",
}


class TestPackageManagerServicePatch(BaseAEATestCase):
    """Test package manager V0 service patch."""

    def test_update_dependencies(self) -> None:
        """Test update dependencies."""

        pm = PackageManagerV1(path=self.packages_dir_path)

        updated_config = SERVICE_CONFIG.copy()
        updated_config[
            "agent"
        ] = "valory/hello_world:0.1.0:bafybei0000000000000000000000000000000000000000000000000001"

        with mock.patch(
            "autonomy.cli.packages.load_yaml",
            return_value=[updated_config, None],
        ), mock.patch(
            "autonomy.cli.packages.dump_yaml"
        ) as dump_yaml_patch, mock.patch.object(
            pm, "get_package_hash", return_value=DUMMY_HASH
        ):
            pm.update_dependencies(package_id=SERVICE_ID)
            dump_yaml_patch.assert_called_with(
                file_path=(
                    pm.package_path_from_package_id(package_id=SERVICE_ID)
                    / DEFAULT_SERVICE_CONFIG_FILE
                ),
                data=SERVICE_CONFIG,
                extra_data=None,
            )

    def test_check_dependencies(
        self,
    ) -> None:
        """Test check dependencies method."""

        pm = PackageManagerV1(path=self.packages_dir_path)
        service_config = SERVICE_CONFIG.copy()
        service_config["license_"] = service_config.pop("license")

        with mock.patch.object(pm, "get_package_hash", return_value=DUMMY_HASH):
            errors = pm.check_dependencies(
                configuration=Service(**service_config),  # type: ignore
            )

            assert len(errors) == 0

    def test_check_dependencies_failure(
        self,
    ) -> None:
        """Test check dependencies method."""

        pm = PackageManagerV1(path=self.packages_dir_path)
        service_config = SERVICE_CONFIG.copy()
        service_config["license_"] = service_config.pop("license")
        with mock.patch.object(pm, "get_package_hash", return_value=None):
            errors = pm.check_dependencies(
                configuration=Service(**service_config),  # type: ignore
            )

            assert len(errors) == 1

            ((_, error),) = errors
            assert error == DepedencyMismatchErrors.HASH_NOT_FOUND

        service_config[
            "agent"
        ] = "valory/hello_world:0.1.0:bafybei0000000000000000000000000000000000000000000000000001"

        with mock.patch.object(pm, "get_package_hash", return_value=DUMMY_HASH):
            errors = pm.check_dependencies(
                configuration=Service(**service_config),  # type: ignore
            )

            assert len(errors) == 1

            ((_, error),) = errors
            assert error == DepedencyMismatchErrors.HASH_DOES_NOT_MATCH


def test_get_package_manager() -> None:
    """Test `get_package_manager` method."""

    with tempfile.TemporaryDirectory() as temp_dir:
        packages_dir = Path(temp_dir)
        (packages_dir / "packages.json").write_text(
            json.dumps({"service/valory/service/0.1.0": DUMMY_HASH})
        )

        pm_v0 = get_package_manager(packages_dir=packages_dir)
        assert isinstance(pm_v0, PackageManagerV0)

    with tempfile.TemporaryDirectory() as temp_dir:
        packages_dir = Path(temp_dir)
        (packages_dir / "packages.json").write_text(
            json.dumps(
                {"dev": {"service/valory/service/0.1.0": DUMMY_HASH}, "third_party": {}}
            )
        )

        pm_v0 = get_package_manager(packages_dir=packages_dir)
        assert isinstance(pm_v0, PackageManagerV1)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the shared state for the price estimation ABCI application."""

import inspect
import json
from abc import ABC, ABCMeta
from collections import Counter
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from time import time
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    OrderedDict,
    Tuple,
    Type,
    cast,
    get_type_hints,
)

from aea.configurations.data_types import PublicId
from aea.exceptions import enforce
from aea.skills.base import Model, SkillContext

from packages.valory.protocols.http.message import HttpMessage
from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppDB,
    BaseSynchronizedData,
    ROUND_COUNT_DEFAULT,
    RoundSequence,
    VALUE_NOT_PROVIDED,
    get_name,
)
from packages.valory.skills.abstract_round_abci.utils import (
    check,
    check_type,
    consensus_threshold,
    get_data_from_nested_dict,
    get_value_with_type,
)


MIN_RESET_PAUSE_DURATION = 10
NUMBER_OF_RETRIES: int = 5
DEFAULT_BACKOFF_FACTOR: float = 2.0
DEFAULT_TYPE_NAME: str = "str"


class FrozenMixin:  # pylint: disable=too-few-public-methods
    """Mixin for classes to enforce read-only attributes."""

    _frozen: bool = False

    def __delattr__(self, *args: Any) -> None:
        """Override __delattr__ to make object immutable."""
        if self._frozen:
            raise AttributeError(
                "This object is frozen! To unfreeze switch `self._frozen` via `__dict__`."
            )
        super().__delattr__(*args)

    def __setattr__(self, *args: Any) -> None:
        """Override __setattr__ to make object immutable."""
        if self._frozen:
            raise AttributeError(
                "This object is frozen! To unfreeze switch `self._frozen` via `__dict__`."
            )
        super().__setattr__(*args)


class TypeCheckMixin:  # pylint: disable=too-few-public-methods
    """Mixin for data classes & models to enforce attribute types on construction."""

    def __post_init__(self) -> None:
        """Check that the type of the provided attributes is correct."""
        for attr, type_ in get_type_hints(self).items():
            value = getattr(self, attr)
            check_type(attr, value, type_)

    @classmethod
    def _ensure(cls, key: str, kwargs: Dict, type_: Any) -> Any:
        """Get and ensure the configuration field is not None (if no default is provided) and of correct type."""
        enforce("skill_context" in kwargs, "Only use on models!")
        skill_id = kwargs["skill_context"].skill_id
        enforce(
            key in kwargs,
            f"'{key}' of type '{type_}' required, but it is not set in `models.params.args` of `skill.yaml` of `{skill_id}`",
        )
        value = kwargs.pop(key)
        try:
            check_type(key, value, type_)
        except TypeError:  # pragma: nocover
            enforce(
                False,
                f"'{key}' must be a {type_}, but type {type(value)} was found in `models.params.args` of `skill.yaml` of `{skill_id}`",
            )
        return value


@dataclass(frozen=True)
class GenesisBlock(TypeCheckMixin):
    """A dataclass to store the genesis block."""

    max_bytes: str
    max_gas: str
    time_iota_ms: str

    def to_json(self) -> Dict[str, str]:
        """Get a GenesisBlock instance as a json dictionary."""
        return {
            "max_bytes": self.max_bytes,
            "max_gas": self.max_gas,
            "time_iota_ms": self.time_iota_ms,
        }


@dataclass(frozen=True)
class GenesisEvidence(TypeCheckMixin):
    """A dataclass to store the genesis evidence."""

    max_age_num_blocks: str
    max_age_duration: str
    max_bytes: str

    def to_json(self) -> Dict[str, str]:
        """Get a GenesisEvidence instance as a json dictionary."""
        return {
            "max_age_num_blocks": self.max_age_num_blocks,
            "max_age_duration": self.max_age_duration,
            "max_bytes": self.max_bytes,
        }


@dataclass(frozen=True)
class GenesisValidator(TypeCheckMixin):
    """A dataclass to store the genesis validator."""

    pub_key_types: Tuple[str, ...]

    def to_json(self) -> Dict[str, List[str]]:
        """Get a GenesisValidator instance as a json dictionary."""
        return {"pub_key_types": list(self.pub_key_types)}


@dataclass(frozen=True)
class GenesisConsensusParams(TypeCheckMixin):
    """A dataclass to store the genesis consensus parameters."""

    block: GenesisBlock
    evidence: GenesisEvidence
    validator: GenesisValidator
    version: dict

    @classmethod
    def from_json_dict(cls, json_dict: dict) -> "GenesisConsensusParams":
        """Get a GenesisConsensusParams instance from a json dictionary."""
        block = GenesisBlock(**json_dict["block"])
        evidence = GenesisEvidence(**json_dict["evidence"])
        validator = GenesisValidator(tuple(json_dict["validator"]["pub_key_types"]))
        return cls(block, evidence, validator, json_dict["version"])

    def to_json(self) -> Dict[str, Any]:
        """Get a GenesisConsensusParams instance as a json dictionary."""
        return {
            "block": self.block.to_json(),
            "evidence": self.evidence.to_json(),
            "validator": self.validator.to_json(),
            "version": self.version,
        }


@dataclass(frozen=True)
class GenesisConfig(TypeCheckMixin):
    """A dataclass to store the genesis configuration."""

    genesis_time: str
    chain_id: str
    consensus_params: GenesisConsensusParams
    voting_power: str

    @classmethod
    def from_json_dict(cls, json_dict: dict) -> "GenesisConfig":
        """Get a GenesisConfig instance from a json dictionary."""
        consensus_params = GenesisConsensusParams.from_json_dict(
            json_dict["consensus_params"]
        )
        return cls(
            json_dict["genesis_time"],
            json_dict["chain_id"],
            consensus_params,
            json_dict["voting_power"],
        )

    def to_json(self) -> Dict[str, Any]:
        """Get a GenesisConfig instance as a json dictionary."""
        return {
            "genesis_time": self.genesis_time,
            "chain_id": self.chain_id,
            "consensus_params": self.consensus_params.to_json(),
            "voting_power": self.voting_power,
        }


class BaseParams(
    Model, FrozenMixin, TypeCheckMixin
):  # pylint: disable=too-many-instance-attributes
    """Parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the parameters object.

        The genesis configuration should be a dictionary with the following format:
            genesis_time: str
            chain_id: str
            consensus_params:
              block:
                max_bytes: str
                max_gas: str
                time_iota_ms: str
              evidence:
                max_age_num_blocks: str
                max_age_duration: str
                max_bytes: str
              validator:
                pub_key_types: List[str]
              version: dict
            voting_power: str

        :param args: positional arguments
        :param kwargs: keyword arguments
        """
        self.genesis_config: GenesisConfig = GenesisConfig.from_json_dict(
            self._ensure("genesis_config", kwargs, dict)
        )
        self.service_id: str = self._ensure("service_id", kwargs, str)
        self.tendermint_url: str = self._ensure("tendermint_url", kwargs, str)
        self.max_healthcheck: int = self._ensure("max_healthcheck", kwargs, int)
        self.round_timeout_seconds: float = self._ensure(
            "round_timeout_seconds", kwargs, float
        )
        self.sleep_time: int = self._ensure("sleep_time", kwargs, int)
        self.retry_timeout: int = self._ensure("retry_timeout", kwargs, int)
        self.retry_attempts: int = self._ensure("retry_attempts", kwargs, int)
        self.keeper_timeout: float = self._ensure("keeper_timeout", kwargs, float)
        self.reset_pause_duration: int = self._ensure_gte(
            "reset_pause_duration", kwargs, int, min_value=MIN_RESET_PAUSE_DURATION
        )
        self.drand_public_key: str = self._ensure("drand_public_key", kwargs, str)
        self.tendermint_com_url: str = self._ensure("tendermint_com_url", kwargs, str)
        self.tendermint_max_retries: int = self._ensure(
            "tendermint_max_retries", kwargs, int
        )
        self.tendermint_check_sleep_delay: int = self._ensure(
            "tendermint_check_sleep_delay", kwargs, int
        )
        self.reset_tendermint_after: int = self._ensure(
            "reset_tendermint_after", kwargs, int
        )
        self.cleanup_history_depth: int = self._ensure(
            "cleanup_history_depth", kwargs, int
        )
        self.cleanup_history_depth_current: Optional[int] = self._ensure(
            "cleanup_history_depth_current", kwargs, Optional[int]
        )
        self.request_timeout: float = self._ensure("request_timeout", kwargs, float)
        self.request_retry_delay: float = self._ensure(
            "request_retry_delay", kwargs, float
        )
        self.tx_timeout: float = self._ensure("tx_timeout", kwargs, float)
        self.max_attempts: int = self._ensure("max_attempts", kwargs, int)
        self.service_registry_address: Optional[str] = self._ensure(
            "service_registry_address", kwargs, Optional[str]
        )
        self.on_chain_service_id: Optional[int] = self._ensure(
            "on_chain_service_id", kwargs, Optional[int]
        )
        self.share_tm_config_on_startup: bool = self._ensure(
            "share_tm_config_on_startup", kwargs, bool
        )
        self.tendermint_p2p_url: str = self._ensure("tendermint_p2p_url", kwargs, str)
        self.use_termination: bool = self._ensure("use_termination", kwargs, bool)
        self.setup_params: Dict[str, Any] = self._ensure("setup", kwargs, dict)

        # we sanitize for null values as these are just kept for schema definitions
        skill_id = kwargs["skill_context"].skill_id
        super().__init__(*args, **kwargs)

        if not self.context.is_abstract_component:
            # setup data are mandatory for non-abstract skills,
            # and they should always contain at least `all_participants` and `safe_contract_address`
            self._ensure_setup(
                {
                    get_name(BaseSynchronizedData.safe_contract_address): str,
                    get_name(BaseSynchronizedData.all_participants): List[str],
                    get_name(BaseSynchronizedData.consensus_threshold): cast(
                        Type, Optional[int]
                    ),
                },
                skill_id,
            )
        self._frozen = True

    def _ensure_setup(
        self, necessary_params: Dict[str, Type], skill_id: PublicId
    ) -> Any:
        """Ensure that the `setup` params contain all the `necessary_keys` and have the correct types."""
        enforce(bool(self.setup_params), "`setup` params contain no values!")

        for key, type_ in necessary_params.items():
            # check that the key is present, note that None is acceptable for optional keys
            value = self.setup_params.get(key, VALUE_NOT_PROVIDED)
            if value is VALUE_NOT_PROVIDED:
                fail_msg = f"Value for `{key}` missing from the `setup` params."
                enforce(False, fail_msg)

            # check that the value is of the correct type
            try:
                check_type(key, value, type_)
            except TypeError:  # pragma: nocover
                enforce(
                    False,
                    f"'{key}' must be a {type_}, but type {type(value)} was found in `models.params.args.setup` "
                    f"of `skill.yaml` of `{skill_id}`",
                )

    def _ensure_gte(
        self, key: str, kwargs: Dict[str, Any], type_: Type, min_value: Any
    ) -> Any:
        """Ensure that the value for the key is greater than or equal to the provided min_value."""
        err = check(min_value, type_)
        enforce(
            err is None,
            f"min_value must be of type {type_.__name__}, but got {type(min_value).__name__}.",
        )
        value = self._ensure(key, kwargs, type_)
        enforce(
            value >= min_value, f"`{key}` must be greater than or equal to {min_value}."
        )
        return value


class _MetaSharedState(ABCMeta):
    """A metaclass that validates SharedState's attributes."""

    def __new__(mcs, name: str, bases: Tuple, namespace: Dict, **kwargs: Any) -> Type:  # type: ignore
        """Initialize the class."""
        new_cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        if ABC in bases:
            # abstract class, return
            return new_cls
        if not issubclass(new_cls, SharedState):
            # the check only applies to SharedState subclasses
            return new_cls

        mcs._check_consistency(cast(Type[SharedState], new_cls))
        return new_cls

    @classmethod
    def _check_consistency(mcs, shared_state_cls: Type["SharedState"]) -> None:
        """Check consistency of class attributes."""
        mcs._check_required_class_attributes(shared_state_cls)

    @classmethod
    def _check_required_class_attributes(
        mcs, shared_state_cls: Type["SharedState"]
    ) -> None:
        """Check that required class attributes are set."""
        if not hasattr(shared_state_cls, "abci_app_cls"):
            raise AttributeError(f"'abci_app_cls' not set on {shared_state_cls}")
        abci_app_cls = shared_state_cls.abci_app_cls
        if not inspect.isclass(abci_app_cls):
            raise AttributeError(f"The object `{abci_app_cls}` is not a class")
        if not issubclass(abci_app_cls, AbciApp):
            cls_name = AbciApp.__name__
            cls_module = AbciApp.__module__
            raise AttributeError(
                f"The class {abci_app_cls} is not an instance of {cls_module}.{cls_name}"
            )


class SharedState(Model, ABC, metaclass=_MetaSharedState):  # type: ignore
    """Keep the current shared state of the skill."""

    abci_app_cls: Type[AbciApp]

    def __init__(
        self,
        *args: Any,
        skill_context: SkillContext,
        **kwargs: Any,
    ) -> None:
        """Initialize the state."""
        self.abci_app_cls._is_abstract = skill_context.is_abstract_component
        self._round_sequence: Optional[RoundSequence] = None
        # a mapping of the agents' addresses to their initial Tendermint configuration, to be retrieved via ACN
        self.initial_tm_configs: Dict[str, Optional[Dict[str, Any]]] = {}
        # a mapping of the other agents' addresses to ACN deliverables
        self.address_to_acn_deliverable: Dict[str, Any] = {}
        self.tm_recovery_params: TendermintRecoveryParams = TendermintRecoveryParams(
            self.abci_app_cls.initial_round_cls.auto_round_id()
        )
        kwargs["skill_context"] = skill_context
        super().__init__(*args, **kwargs)

    def acn_container(self) -> Dict[str, Any]:
        """Create a container for ACN results, i.e., a mapping from others' addresses to `None`."""
        ourself = {self.context.agent_address}
        others_addresses = self.synchronized_data.all_participants - ourself

        return dict.fromkeys(others_addresses)

    def setup(self) -> None:
        """Set up the model."""
        self._round_sequence = RoundSequence(self.abci_app_cls)
        setup_params = cast(BaseParams, self.context.params).setup_params
        self.round_sequence.setup(
            BaseSynchronizedData(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(setup_params),
                    cross_period_persisted_keys=self.abci_app_cls.cross_period_persisted_keys,
                )
            ),
            self.context.logger,
        )
        if not self.context.is_abstract_component:
            self.initial_tm_configs = dict.fromkeys(
                self.synchronized_data.all_participants
            )

    @property
    def round_sequence(self) -> RoundSequence:
        """Get the round_sequence."""
        if self._round_sequence is None:
            raise ValueError("round sequence not available")
        return self._round_sequence

    @property
    def synchronized_data(self) -> BaseSynchronizedData:
        """Get the latest synchronized_data if available."""
        return self.round_sequence.latest_synchronized_data

    def get_acn_result(self) -> Any:
        """Get the majority of the ACN deliverables."""
        if len(self.address_to_acn_deliverable) == 0:
            return None

        # the current agent does not participate, so we need `nb_participants - 1`
        threshold = consensus_threshold(self.synchronized_data.nb_participants - 1)
        counter = Counter(self.address_to_acn_deliverable.values())
        most_common_value, n_appearances = counter.most_common(1)[0]

        if n_appearances < threshold:
            return None

        self.context.logger.debug(
            f"ACN result is '{most_common_value}' from '{self.address_to_acn_deliverable}'."
        )
        return most_common_value


class Requests(Model, FrozenMixin):
    """Keep the current pending requests."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the state."""
        # mapping from dialogue reference nonce to a callback
        self.request_id_to_callback: Dict[str, Callable] = {}
        super().__init__(*args, **kwargs)
        self._frozen = True


class UnexpectedResponseError(Exception):
    """Exception class for unexpected responses from Apis."""


@dataclass
class ResponseInfo(TypeCheckMixin):
    """A dataclass to hold all the information related to the response."""

    response_key: Optional[str]
    response_index: Optional[int]
    response_type: str
    error_key: Optional[str]
    error_index: Optional[int]
    error_type: str
    error_data: Any = None

    @classmethod
    def from_json_dict(cls, kwargs: Dict) -> "ResponseInfo":
        """Initialize a response info object from kwargs."""
        response_key: Optional[str] = kwargs.pop("response_key", None)
        response_index: Optional[int] = kwargs.pop("response_index", None)
        response_type: str = kwargs.pop("response_type", DEFAULT_TYPE_NAME)
        error_key: Optional[str] = kwargs.pop("error_key", None)
        error_index: Optional[int] = kwargs.pop("error_index", None)
        error_type: str = kwargs.pop("error_type", DEFAULT_TYPE_NAME)
        return cls(
            response_key,
            response_index,
            response_type,
            error_key,
            error_index,
            error_type,
        )


@dataclass
class RetriesInfo(TypeCheckMixin):
    """A dataclass to hold all the information related to the retries."""

    retries: int
    backoff_factor: float
    retries_attempted: int = 0

    @classmethod
    def from_json_dict(cls, kwargs: Dict) -> "RetriesInfo":
        """Initialize a retries info object from kwargs."""
        retries: int = kwargs.pop("retries", NUMBER_OF_RETRIES)
        backoff_factor: float = kwargs.pop("backoff_factor", DEFAULT_BACKOFF_FACTOR)
        return cls(retries, backoff_factor)

    @property
    def suggested_sleep_time(self) -> float:
        """The suggested amount of time to sleep."""
        return self.backoff_factor**self.retries_attempted


@dataclass(frozen=True)
class TendermintRecoveryParams(TypeCheckMixin):
    """
    A dataclass to hold all parameters related to agent <-> tendermint recovery procedures.

    This must be frozen so that we make sure it does not get edited.
    """

    reset_from_round: str
    round_count: int = ROUND_COUNT_DEFAULT
    reset_params: Optional[Dict[str, str]] = None
    serialized_db_state: Optional[str] = None

    def __hash__(self) -> int:
        """Hash the object."""
        return hash(
            self.reset_from_round
            + str(self.round_count)
            + str(self.serialized_db_state)
            + json.dumps(self.reset_params, sort_keys=True)
        )


class ApiSpecs(Model, FrozenMixin, TypeCheckMixin):
    """A model that wraps APIs to get cryptocurrency prices."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize ApiSpecsModel."""
        self.url: str = self._ensure("url", kwargs, str)
        self.api_id: str = self._ensure("api_id", kwargs, str)
        self.method: str = self._ensure("method", kwargs, str)
        self.headers: Dict[str, str] = dict(
            self._ensure("headers", kwargs, OrderedDict[str, str])
        )
        self.parameters: Dict[str, str] = dict(
            self._ensure("parameters", kwargs, OrderedDict[str, str])
        )
        self.response_info = ResponseInfo.from_json_dict(kwargs)
        self.retries_info = RetriesInfo.from_json_dict(kwargs)
        super().__init__(*args, **kwargs)
        self._frozen = True

    def get_spec(
        self,
    ) -> Dict:
        """Returns dictionary containing api specifications."""

        return {
            "url": self.url,
            "method": self.method,
            "headers": self.headers,
            "parameters": self.parameters,
        }

    def _log_response(self, decoded_response: str) -> None:
        """Log the decoded response message using error level."""
        pretty_json_str = json.dumps(decoded_response, indent=4)
        self.context.logger.error(f"Response: {pretty_json_str}")

    @staticmethod
    def _parse_response(
        response_data: Any,
        response_keys: Optional[str],
        response_index: Optional[int],
        response_type: str,
    ) -> Any:
        """Parse a response from an API."""
        if response_keys is not None:
            response_data = get_data_from_nested_dict(response_data, response_keys)

        if response_index is not None:
            response_data = response_data[response_index]

        return get_value_with_type(response_data, response_type)

    def _get_error_from_response(self, response_data: Any) -> Any:
        """Try to get an error from the response."""
        try:
            return self._parse_response(
                response_data,
                self.response_info.error_key,
                self.response_info.error_index,
                self.response_info.error_type,
            )
        except (KeyError, IndexError, TypeError):
            self.context.logger.error(
                f"Could not parse error using the given key(s) ({self.response_info.error_key}) "
                f"and index ({self.response_info.error_index})!"
            )
            return None

    def _parse_response_data(self, response_data: Any) -> Any:
        """Get the response data."""
        try:
            return self._parse_response(
                response_data,
                self.response_info.response_key,
                self.response_info.response_index,
                self.response_info.response_type,
            )
        except (KeyError, IndexError, TypeError) as e:
            raise UnexpectedResponseError from e

    def process_response(self, response: HttpMessage) -> Any:
        """Process response from api."""
        decoded_response = response.body.decode()
        self.response_info.error_data = None

        try:
            response_data = json.loads(decoded_response)
        except json.JSONDecodeError:
            self.context.logger.error("Could not parse the response body!")
            self._log_response(decoded_response)
            return None

        try:
            return self._parse_response_data(response_data)
        except UnexpectedResponseError:
            self.context.logger.error(
                f"Could not access response using the given key(s) ({self.response_info.response_key}) "
                f"and index ({self.response_info.response_index})!"
            )
            self._log_response(decoded_response)
            self.response_info.error_data = self._get_error_from_response(response_data)
            return None

    def increment_retries(self) -> None:
        """Increment the retries counter."""
        self.retries_info.retries_attempted += 1

    def reset_retries(self) -> None:
        """Reset the retries counter."""
        self.retries_info.retries_attempted = 0

    def is_retries_exceeded(self) -> bool:
        """Check if the retries amount has been exceeded."""
        return self.retries_info.retries_attempted > self.retries_info.retries


class BenchmarkBlockTypes(Enum):
    """Benchmark block types."""

    LOCAL = "local"
    CONSENSUS = "consensus"
    TOTAL = "total"


class BenchmarkBlock:
    """
    Benchmark

    This class represents logic to measure the code block using a
    context manager.
    """

    start: float
    total_time: float
    block_type: str

    def __init__(self, block_type: str) -> None:
        """Benchmark for single round."""
        self.block_type = block_type
        self.start = 0
        self.total_time = 0

    def __enter__(
        self,
    ) -> None:
        """Enter context."""
        self.start = time()

    def __exit__(self, *args: List, **kwargs: Dict) -> None:
        """Exit context"""
        self.total_time = time() - self.start


class BenchmarkBehaviour:
    """
    BenchmarkBehaviour

    This class represents logic to benchmark a single behaviour.
    """

    local_data: Dict[str, BenchmarkBlock]

    def __init__(
        self,
    ) -> None:
        """Initialize Benchmark behaviour object."""
        self.local_data = {}

    def _measure(self, block_type: str) -> BenchmarkBlock:
        """
        Returns a BenchmarkBlock object.

        :param block_type: type of block (e.g. local, consensus, request)
        :return: BenchmarkBlock
        """

        if block_type not in self.local_data:
            self.local_data[block_type] = BenchmarkBlock(block_type)

        return self.local_data[block_type]

    def local(
        self,
    ) -> BenchmarkBlock:
        """Measure local block."""
        return self._measure(BenchmarkBlockTypes.LOCAL.value)

    def consensus(
        self,
    ) -> BenchmarkBlock:
        """Measure consensus block."""
        return self._measure(BenchmarkBlockTypes.CONSENSUS.value)


class BenchmarkTool(Model, TypeCheckMixin, FrozenMixin):
    """
    BenchmarkTool

    Tool to benchmark ABCI apps.
    """

    benchmark_data: Dict[str, BenchmarkBehaviour]
    log_dir: Path

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Benchmark tool for rounds behaviours."""
        self.benchmark_data = {}
        log_dir_ = self._ensure("log_dir", kwargs, str)
        self.log_dir = Path(log_dir_)
        super().__init__(*args, **kwargs)
        self._frozen = True

    def measure(self, behaviour: str) -> BenchmarkBehaviour:
        """Measure time to complete round."""
        if behaviour not in self.benchmark_data:
            self.benchmark_data[behaviour] = BenchmarkBehaviour()
        return self.benchmark_data[behaviour]

    @property
    def data(
        self,
    ) -> List:
        """Returns formatted data."""

        behavioural_data = []
        for behaviour, tool in self.benchmark_data.items():
            data = {k: v.total_time for k, v in tool.local_data.items()}
            data[BenchmarkBlockTypes.TOTAL.value] = sum(data.values())
            behavioural_data.append({"behaviour": behaviour, "data": data})

        return behavioural_data

    def save(self, period: int = 0, reset: bool = True) -> None:
        """Save logs to a file."""

        try:
            self.log_dir.mkdir(exist_ok=True)
            agent_dir = self.log_dir / self.context.agent_address
            agent_dir.mkdir(exist_ok=True)
            filepath = agent_dir / f"{period}.json"

            with open(str(filepath), "w+", encoding="utf-8") as outfile:
                json.dump(self.data, outfile)
            self.context.logger.info(f"Saving benchmarking data for period: {period}")

        except PermissionError as e:  # pragma: nocover
            self.context.logger.info(f"Error saving benchmark data:\n{e}")

        if reset:
            self.reset()

    def reset(
        self,
    ) -> None:
        """Reset benchmark data"""
        self.benchmark_data.clear()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Docker-compose Deployment Generator."""
import ipaddress
from pathlib import Path
from typing import Dict, Optional, cast

from aea.configurations.constants import DEFAULT_LEDGER, PRIVATE_KEY_PATH_SCHEMA
from docker import from_env

from autonomy.constants import (
    ACN_IMAGE_NAME,
    ACN_IMAGE_VERSION,
    DOCKER_COMPOSE_YAML,
    HARDHAT_IMAGE_NAME,
    HARDHAT_IMAGE_VERSION,
    OAR_IMAGE,
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)
from autonomy.deploy.base import BaseDeploymentGenerator
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA,
    DEPLOYMENT_KEY_DIRECTORY,
    INFO,
    KEY_SCHEMA_PRIVATE_KEY,
    KEY_SCHEMA_TYPE,
)
from autonomy.deploy.generators.docker_compose.templates import (
    ABCI_NODE_TEMPLATE,
    ACN_NODE_TEMPLATE,
    DOCKER_COMPOSE_TEMPLATE,
    HARDHAT_NODE_TEMPLATE,
    PORTS,
    PORT_MAPPING_CONFIG,
    TENDERMINT_CONFIG_TEMPLATE,
    TENDERMINT_NODE_TEMPLATE,
)


NETWORK_ADDRESS_OFFSET = 1
BASE_SUBNET = cast(ipaddress.IPv4Network, ipaddress.ip_network("192.167.11.0/24"))
SUBNET_OVERFLOW = 256

DEFAULT_PACKAGES_PATH = Path.cwd().absolute() / "packages"
DEFAULT_OPEN_AEA_DIR: Path = Path.home().absolute() / "open-aea"
DEFAULT_OPEN_AUTONOMY_DIR: Path = Path.home().absolute() / "open-autonomy"


def build_tendermint_node_config(  # pylint: disable=too-many-arguments
    node_id: int,
    container_name: str,
    abci_node: str,
    network_name: str,
    network_address: str,
    dev_mode: bool = False,
    log_level: str = INFO,
    tendermint_ports: Optional[Dict[int, int]] = None,
) -> str:
    """Build tendermint node config for docker compose."""

    config = TENDERMINT_NODE_TEMPLATE.format(
        node_id=node_id,
        container_name=container_name,
        abci_node=abci_node,
        network_address=network_address,
        localnet_port_range=node_id,
        log_level=log_level,
        tendermint_image_name=TENDERMINT_IMAGE_NAME,
        tendermint_image_version=TENDERMINT_IMAGE_VERSION,
        network_name=network_name,
    )

    if dev_mode:
        config += "      - ./persistent_data/tm_state:/tm_state:Z"
        config = config.replace("DEV_MODE=0", "DEV_MODE=1")

    if tendermint_ports is not None:
        port_mappings = map(
            lambda x: PORT_MAPPING_CONFIG.format(host_port=x[0], container_port=x[1]),
            tendermint_ports.items(),
        )
        port_config = "\n".join([PORTS, *port_mappings])
        config += port_config
        config += "\n"

    return config


def build_agent_config(  # pylint: disable=too-many-arguments
    node_id: int,
    container_name: str,
    agent_vars: Dict,
    runtime_image: str,
    network_name: str,
    network_address: str,
    dev_mode: bool = False,
    package_dir: Path = DEFAULT_PACKAGES_PATH,
    open_aea_dir: Path = DEFAULT_OPEN_AEA_DIR,
    open_autonomy_dir: Path = DEFAULT_OPEN_AUTONOMY_DIR,
    agent_ports: Optional[Dict[int, int]] = None,
) -> str:
    """Build agent config."""

    agent_vars_string = "\n".join([f"      - {k}={v}" for k, v in agent_vars.items()])
    config = ABCI_NODE_TEMPLATE.format(
        node_id=node_id,
        container_name=container_name,
        agent_vars=agent_vars_string,
        network_address=network_address,
        runtime_image=runtime_image,
        network_name=network_name,
    )

    if dev_mode:
        config += "      - ./persistent_data/benchmarks:/benchmarks:Z\n"
        config += (
            "      - ./persistent_data/venvs:/home/ubuntu/.local/share/virtualenvs:Z\n"
        )
        config += f"      - {package_dir}:/home/ubuntu/packages:rw\n"
        config += f"      - {open_aea_dir}:/open-aea\n"
        config += f"      - {open_autonomy_dir}:/open-autonomy\n"

    if agent_ports is not None:
        port_mappings = map(
            lambda x: PORT_MAPPING_CONFIG.format(host_port=x[0], container_port=x[1]),
            agent_ports.items(),
        )
        port_config = "\n".join([PORTS, *port_mappings])
        config += port_config
        config += "\n"

    return config


class Network:
    """Class to represent network of the service."""

    def __init__(
        self,
        name: str,
        base: ipaddress.IPv4Network = BASE_SUBNET,
    ) -> None:
        """Initialize."""
        self.name = name
        self.base = base
        self.subnet = self.build()

        self._address_offeset = NETWORK_ADDRESS_OFFSET

    @staticmethod
    def next_subnet(subnet: ipaddress.IPv4Network) -> ipaddress.IPv4Network:
        """Calculat next available subnet."""
        new_address = subnet.network_address + SUBNET_OVERFLOW
        return cast(
            ipaddress.IPv4Network,
            ipaddress.ip_network(f"{new_address}/{subnet.prefixlen}"),
        )

    def build(
        self,
    ) -> ipaddress.IPv4Network:
        """Initialize network params."""
        docker = from_env()
        used_subnets = set()
        for network in docker.networks.list():
            # Network already exists
            if f"abci_build_{self.name}" == network.attrs["Name"]:
                config, *_ = network.attrs["IPAM"]["Config"]
                return cast(
                    ipaddress.IPv4Network, ipaddress.ip_network(config["Subnet"])
                )
            for config in network.attrs["IPAM"]["Config"]:
                used_subnets.add(config["Subnet"])

        subnet = self.base
        while str(subnet) in used_subnets:
            subnet = self.next_subnet(subnet=subnet)
        return subnet

    @property
    def next_address(self) -> str:
        """Returns the next IP address string."""
        self._address_offeset += 1
        return str(self.subnet.network_address + self._address_offeset)


class DockerComposeGenerator(BaseDeploymentGenerator):
    """Class to automate the generation of Deployments."""

    output_name: str = DOCKER_COMPOSE_YAML
    deployment_type: str = "docker-compose"

    def generate_config_tendermint(self) -> "DockerComposeGenerator":
        """Generate the command to configure tendermint testnet."""

        if not self.use_tm_testnet_setup:
            return self

        hosts = " ".join(
            [
                "--hostname=" + self.service_builder.get_tm_container_name(index=k)
                for k in range(self.service_builder.service.number_of_agents)
            ]
        )
        self.tendermint_job_config = TENDERMINT_CONFIG_TEMPLATE.format(
            validators=self.service_builder.service.number_of_agents, hosts=hosts
        )
        client = from_env()
        image = f"{TENDERMINT_IMAGE_NAME}:{TENDERMINT_IMAGE_VERSION}"

        run_log = client.containers.run(
            image=image,
            volumes={f"{self.build_dir}/nodes": {"bind": "/tendermint", "mode": "z"}},
            entrypoint=self.tendermint_job_config,
        )
        print(run_log.decode())

        return self

    def generate(
        self,
        image_version: Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "DockerComposeGenerator":
        """Generate the new configuration."""

        network_name = f"service_{self.service_builder.service.name}_localnet"
        network = Network(name=network_name)
        image_version = image_version or self.service_builder.service.agent.hash
        if self.dev_mode:
            image_version = "dev"

        runtime_image = OAR_IMAGE.format(
            image_author=self.image_author,
            agent=self.service_builder.service.agent.name,
            version=image_version,
        )
        agent_vars = self.service_builder.generate_agents()
        agents = "".join(
            [
                build_agent_config(
                    node_id=i,
                    container_name=self.service_builder.get_abci_container_name(
                        index=i
                    ),
                    runtime_image=runtime_image,
                    agent_vars=agent_vars[i],
                    dev_mode=self.dev_mode,
                    package_dir=self.packages_dir,
                    open_aea_dir=self.open_aea_dir,
                    open_autonomy_dir=self.open_autonomy_dir,
                    agent_ports=(
                        self.service_builder.service.deployment_config.get("agent", {})
                        .get("ports", {})
                        .get(i)
                    ),
                    network_name=network_name,
                    network_address=network.next_address,
                )
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )
        tendermint_nodes = "".join(
            [
                build_tendermint_node_config(
                    node_id=i,
                    container_name=self.service_builder.get_tm_container_name(index=i),
                    abci_node=self.service_builder.get_abci_container_name(index=i),
                    dev_mode=self.dev_mode,
                    log_level=self.service_builder.log_level,
                    tendermint_ports=(
                        self.service_builder.service.deployment_config.get(
                            "tendermint", {}
                        )
                        .get("ports", {})
                        .get(i)
                    ),
                    network_name=network_name,
                    network_address=network.next_address,
                )
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )

        hardhat_node = ""
        if use_hardhat:
            hardhat_node = HARDHAT_NODE_TEMPLATE.format(
                hardhat_image_name=HARDHAT_IMAGE_NAME,
                hardhat_image_version=HARDHAT_IMAGE_VERSION,
                network_name=network_name,
                network_address=network.next_address,
            )

        acn_node = ""
        if use_acn:
            acn_node = ACN_NODE_TEMPLATE.format(
                acn_image_name=ACN_IMAGE_NAME,
                acn_image_version=ACN_IMAGE_VERSION,
                network_name=network_name,
                network_address=network.next_address,
            )

        self.output = DOCKER_COMPOSE_TEMPLATE.format(
            abci_nodes=agents,
            tendermint_nodes=tendermint_nodes,
            hardhat_node=hardhat_node,
            acn_node=acn_node,
            network_name=network_name,
            subnet=str(network.subnet),
        )

        return self

    def populate_private_keys(
        self,
    ) -> "DockerComposeGenerator":
        """Populate the private keys to the build directory for docker-compose mapping."""
        keys_dir = self.build_dir / DEPLOYMENT_KEY_DIRECTORY
        for x in range(self.service_builder.service.number_of_agents):
            path = keys_dir / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=x)
            ledger = self.service_builder.keys[x].get(KEY_SCHEMA_TYPE, DEFAULT_LEDGER)
            key = self.service_builder.keys[x][KEY_SCHEMA_PRIVATE_KEY]
            keys_file = path / PRIVATE_KEY_PATH_SCHEMA.format(ledger)
            path.mkdir()
            with keys_file.open(mode="w", encoding=DEFAULT_ENCODING) as f:
                f.write(key)

        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the rounds of the skill."""

from copy import deepcopy
from typing import FrozenSet, cast

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciAppDB,
    TransactionNotValidError,
)
from packages.valory.skills.termination_abci.payloads import BackgroundPayload
from packages.valory.skills.termination_abci.rounds import (
    BackgroundRound,
    Event,
    SynchronizedData,
    TerminationRound,
)


MAX_PARTICIPANTS: int = 4


def get_participants() -> FrozenSet[str]:
    """Participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


class BaseRoundTestClass:  # pylint: disable=too-few-public-methods
    """Base test class for Rounds."""

    synchronized_data: SynchronizedData
    participants: FrozenSet[str]

    def setup(
        self,
    ) -> None:
        """Setup the test class."""

        self.participants = get_participants()
        self.synchronized_data = SynchronizedData(
            AbciAppDB(
                setup_data=dict(
                    participants=[tuple(self.participants)],
                    all_participants=[tuple(self.participants)],
                    consensus_threshold=[3],
                ),
            )
        )


class TestBackgroundRound(BaseRoundTestClass):
    """Tests for BackgroundRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = BackgroundRound(
            synchronized_data=deepcopy(self.synchronized_data),
        )
        payload_data = "0xdata"
        first_payload, *payloads = [
            BackgroundPayload(sender=participant, background_data=payload_data)
            for participant in self.participants
        ]

        test_round.process_payload(first_payload)
        assert test_round.collection == {first_payload.sender: first_payload}
        assert test_round.end_block() is None

        for payload in payloads:
            test_round.process_payload(payload)

        expected_state = self.synchronized_data.update(
            most_voted_tx_hash=payload_data,
            termination_majority_reached=True,
        )

        res = test_round.end_block()
        assert res is not None
        actual_state, event = res

        assert (
            cast(SynchronizedData, actual_state).termination_majority_reached
            == cast(  # pylint: disable=no-member
                SynchronizedData, expected_state
            ).termination_majority_reached
        )

        assert event == Event.TERMINATE

    def test_bad_payloads(self) -> None:
        """Tests the background round when bad payloads are sent."""
        test_round = BackgroundRound(
            synchronized_data=deepcopy(self.synchronized_data),
        )
        payload_data = "0xdata"
        bad_participant = "non_existent"
        bad_participant_payload = BackgroundPayload(
            sender=bad_participant, background_data=payload_data
        )
        first_payload, *_ = [
            BackgroundPayload(sender=participant, background_data=payload_data)
            for participant in self.participants
        ]

        with pytest.raises(
            TransactionNotValidError,
            match=f"{bad_participant} not in list of participants",
        ):
            test_round.check_payload(bad_participant_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=f"{bad_participant} not in list of participants",
        ):
            test_round.process_payload(bad_participant_payload)

        # a valid payload gets sent for the first time and it goes through
        test_round.process_payload(first_payload)

        # a duplicate (valid) payload will not go through
        with pytest.raises(
            TransactionNotValidError,
            match=f"sender {first_payload.sender} has already sent value for round",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=f"sender {first_payload.sender} has already sent value for round",
        ):
            test_round.process_payload(first_payload)


class TestTerminationRound(BaseRoundTestClass):
    """Tests for TerminationRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = TerminationRound(
            synchronized_data=deepcopy(self.synchronized_data),
        )
        res = test_round.end_block()  # pylint: disable=assignment-from-none
        assert res is None
